from __future__ import annotations

import math

import numpy as np
import pytest

from NonlinearTMM import Material, SecondOrderNLTMM


@pytest.fixture()
def tmm():
    wlP1 = 1000e-9
    wlP2 = 1000e-9
    polP1 = "s"
    polP2 = "s"
    polGen = "s"
    Ly = 2e-3
    w0P1 = 4000e-6
    w0P2 = 4000e-6
    pwrP1 = 1.1
    E0P2 = 12.0

    crystalD = 50e-6

    # Define materials
    wlsCrystal = np.array([400e-9, 1100e-9])
    nsCrystal = np.array([1.54, 1.53], dtype=complex)
    prism = Material.Static(1.5)
    crystal = Material(wlsCrystal, nsCrystal)
    dielectric = Material.Static(1.6)
    crystal.chi2.Update(d11=1e-12, d22=1e-12, d33=1e-12)

    # Init SecondOrderNLTMM
    tmm = SecondOrderNLTMM()
    tmm.P1.SetParams(wl=wlP1, pol=polP1, beta=0.0, I0=pwrP1 / (Ly * w0P1))
    tmm.P2.SetParams(wl=wlP2, pol=polP2, beta=0.0, overrideE0=True, E0=E0P2)
    tmm.Gen.SetParams(pol=polGen)

    # Add layers
    tmm.AddLayer(math.inf, prism)
    tmm.AddLayer(crystalD, crystal)
    tmm.AddLayer(math.inf, dielectric)

    # Init waves
    waveP1Params = {
        "waveType": "tukey",
        "pwr": pwrP1,
        "w0": w0P1,
        "Ly": Ly,
        "dynamicMaxXCoef": 5,
        "nPointsInteg": 200,
        "maxPhi": np.radians(10.0),
    }
    waveP2Params = {"waveType": "planewave", "w0": w0P2, "Ly": Ly, "overrideE0": True, "E0": E0P2}
    tmm.P1.wave.SetParams(**waveP1Params)
    tmm.P2.wave.SetParams(**waveP2Params)

    return tmm


@pytest.fixture(params=["tukey", "gaussian", "planewave"])
def waveType(request):
    return request.param


@pytest.fixture(params=["p", "s"])
def polGen(request):
    return request.param


@pytest.fixture(params=["p", "s"])
def polP1(request):
    return request.param


@pytest.fixture(params=["p", "s"])
def polP2(request):
    return request.param


def test_SecondOrderNLTMMWaveSweep(tmm, waveType, polP1, polP2, polGen):
    betas = np.linspace(0.0, 0.99, 10)

    tmm.P1.pol = polP1
    tmm.P2.pol = polP2
    tmm.Gen.pol = polGen
    tmm.P1.wave.waveType = waveType

    sr = tmm.Sweep("beta", betas, betas, outP1=False, outP2=False)
    srW = tmm.WaveSweep("beta", betas, betas, outP1=False, outP2=False)

    np.testing.assert_allclose(srW.Gen.Pr, sr.Gen.Ir * srW.Gen.beamArea, rtol=1e-2)
    np.testing.assert_allclose(srW.Gen.Pt, sr.Gen.It * srW.Gen.beamArea, rtol=6e-2)



from __future__ import annotations

import importlib.metadata
import sys
from pathlib import Path

sys.path.insert(0, str(Path(__file__).resolve().parent.parent))

# -- General configuration ------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
    "sphinx.ext.autosummary",
    "numpydoc",
]

source_suffix = ".rst"
root_doc = "index"

project = "NonlinearTMM"
copyright = "2017-2026, Ardi Loot"
author = "Ardi Loot"

__version__ = importlib.metadata.version("NonlinearTMM")
version = __version__
release = __version__

language = "en"
exclude_patterns = ["_build", "_autosummary", "Thumbs.db", ".DS_Store"]
todo_include_todos = True

# Suppress :any: cross-reference warnings from numpydoc — these are parameter
# names (e.g. paramStr, layerNr) that numpydoc converts to :any: roles but
# that have no matching Sphinx targets.
suppress_warnings = ["ref.any"]

# -- autosummary -----------------------------------------------------------

autosummary_generate = True

# -- numpydoc --------------------------------------------------------------

numpydoc_xref_param_type = False
numpydoc_show_class_members = False

# -- intersphinx -----------------------------------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
}

# -- Options for HTML output ----------------------------------------------

html_theme = "furo"
html_theme_options = {
    "navigation_with_keys": True,
}


# -- Options for HTMLHelp output ------------------------------------------

htmlhelp_basename = "NonlinearTMMdoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}

latex_documents = [
    (root_doc, "NonlinearTMM.tex", "NonlinearTMM Documentation", "Ardi Loot", "manual"),
]


# -- Options for manual page output ---------------------------------------

man_pages = [(root_doc, "nonlineartmm", "NonlinearTMM Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

texinfo_documents = [
    (
        root_doc,
        "NonlinearTMM",
        "NonlinearTMM Documentation",
        author,
        "NonlinearTMM",
        "Nonlinear transfer-matrix method.",
        "Miscellaneous",
    ),
]



from __future__ import annotations

import math
import os
from bisect import bisect_left, bisect_right
from pathlib import Path
from typing import Any

import numpy as np
import yaml
from numpy.typing import ArrayLike, NDArray
from scipy.interpolate import interp1d

from LabPy import Core

materialsDir = Path(__file__).resolve().parent / "../materials"
if not materialsDir.is_dir():
    materialsDir = Path(os.environ["PYLAB_MATERIALS_DIR"])

# ===============================================================================
# Methods
# ===============================================================================


def literal_unicode_representer(dumper: yaml.Dumper, data: str) -> Any:
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")


def GetMaterialsList() -> list[str]:
    """Generates list of all materials in database.

    Returns:
        list: list of paths to all available materials

    """
    res = ["Static"]
    res.extend(str(p.relative_to(materialsDir).with_suffix("")) for p in sorted(materialsDir.rglob("*.yml")))
    return res


def SaveMaterialFile(
    filename: str | Path,
    wls: NDArray[np.floating],
    n: NDArray[np.complexfloating],
) -> None:
    with open(filename, "w") as stream:
        data: dict[str, Any] = {}
        data["BOOK"] = "a"
        data["BOOK LONG"] = "b"
        data["PAGE"] = "c"
        data["PAGE LONG"] = "d"
        data["REFERENCES"] = "e"
        data["COMMENTS"] = "f"

        nkData = "\n".join(f"{1e6 * wl_val} {n_val.real} {n_val.imag}" for wl_val, n_val in zip(wls, n))

        class literal_unicode(str):
            pass

        yaml.add_representer(literal_unicode, literal_unicode_representer)
        data["DATA"] = {"type": "nk", "data": literal_unicode(nkData)}

        yaml.dump(data, stream, default_flow_style=False)


def MaterialFromConf(conf: tuple[str, dict[str, Any]]) -> Material:
    param, kwargParams = conf
    return Material(param, **kwargParams)


# ===============================================================================
# Classes
# ===============================================================================


class Material(Core.ParamsBaseClass):
    """This class reads and processes the refractiveindex.info database.

    Args:
        materialFile (str): material filename or Static

    Attributes:
        name (str): The name of material
        longName (str): The long name of material
        page (str): The page of material
        longPage (str): The long name of page of material

    """

    def __init__(self, materialFile: str, boundsError: bool = True, **kwargs: Any) -> None:
        self._params = ["kAdditional"]
        self.materialFile = materialFile
        self.wlExp = None
        self.nExp = None
        self.kExp = None
        self.kAdditional = 0.0
        self.boundsError = boundsError
        self.isFormula = False
        if materialFile == "Static":
            self.n = 1.0
            self.k = 0.0
            self.name = "Static"
            self.nFunc = lambda wl: np.ones_like(wl) * self.n
            self.kFunc = lambda wl: np.ones_like(wl) * self.k
            self._params += ["n", "k"]
        else:
            self._LoadFromFile(materialFile)

        super().__init__(**kwargs)

    def _LoadFromFile(self, materialFile: str) -> None:
        self.filename = materialsDir / f"{materialFile}.yml"

        with open(self.filename) as stream:
            self.rawData = yaml.safe_load(stream)

        self.nFunc = None
        self.kFunc = lambda wl: 0.0

        self.name = self.rawData.get("BOOK", None)
        self.longName = self.rawData.get("BOOK LONG", None)
        self.page = self.rawData.get("PAGE", None)
        self.longPage = self.rawData.get("PAGE LONG", None)
        self.comments = self.rawData.get("COMMENTS", None)
        self.references = self.rawData.get("REFERENCES", None)

        self._ReadDatapoints()
        self._ReadFormulas()

    def __call__(self, wl: ArrayLike) -> NDArray[np.complexfloating]:
        """Returns (interpolates if needed) complex refractive index at wavelength wl.

        Args:
            wl (float): wavelength of light

        Returns:
            complex: complex refractive index

        """
        res = self.nFunc(wl) + 1.0j * self.kFunc(wl) + 1.0j * self.kAdditional
        return res

    def GetN(self, wlFr: float = -math.inf, wlTo: float = -math.inf) -> tuple[NDArray, NDArray]:
        """Returns complex refractive index datapoints in range wlFr - wlTo.

        Args:
           wlFr (float, optional):  The minimum wavelength.
           wlTo (float, optional):  The maximum wavelength.

        Returns:
           tuple_of_arrays:  wavelength and complex refractive index arrays

        """
        index1 = bisect_left(self.wlExp, wlFr)
        index2 = bisect_right(self.wlExp, wlTo)

        wls = self.wlExp[index1:index2]
        n = self.nExp[index1:index2] + 1.0j * self.kExp[index1:index2]
        return wls, n

    # Private methods

    def _ReadDatapoints(self) -> None:
        if "DATA" not in self.rawData or "data" not in self.rawData["DATA"]:
            return

        dataType = self.rawData["DATA"]["type"]
        dataStr = self.rawData["DATA"]["data"].strip().split("\n")

        if dataType == "nk":
            data = np.zeros((len(dataStr), 3))
        elif dataType in ("n", "k"):
            data = np.zeros((len(dataStr), 2))
        else:
            raise ValueError("Unknown data type.")

        for i, line in enumerate(dataStr):
            data[i, :] = np.array(list(map(float, line.split())))

        match dataType:
            case "nk":
                self.wlExp, self.nExp, self.kExp = data.T
            case "n":
                self.wlExp, self.nExp = data.T
            case "k":
                self.wlExp, self.kExp = data.T

        # Convert wavelength to SI units
        self.wlExp *= 1e-6

        # Define interpolation functions
        if self.nExp is not None:
            self.nFunc = interp1d(self.wlExp, self.nExp, bounds_error=self.boundsError)

        if self.kExp is not None:
            self.kFunc = interp1d(self.wlExp, self.kExp, bounds_error=self.boundsError)

    def _ReadFormulas(self) -> None:
        if "FORMULA" in self.rawData:
            dataFormula = self.rawData["FORMULA"]
            wlRange = 1e-6 * np.array(dataFormula["range"].split()).astype(float)
            dispType = dataFormula["type"]
            coefs = np.array(dataFormula["coefficients"].split()).astype(float)
            match dispType:
                case 4:
                    self.nFunc = lambda wl: _DispersionFunc4(wl, coefs)
                case 2:
                    self.nFunc = lambda wl: _DispersionFunc2(wl, coefs)
                case 1:
                    self.nFunc = lambda wl: _DispersionFunc1(wl, coefs)
                case _:
                    raise NotImplementedError(f"Unsupported dispersion type: {dispType}")
            self.isFormula = True
            self.wlRange = wlRange

        if "DATA" in self.rawData and isinstance(self.rawData["DATA"], list):
            coefs = np.array(self.rawData["DATA"][0]["coefficients"].split()).astype(float)
            match self.rawData["DATA"][0]["type"]:
                case "formula 1":
                    self.nFunc = lambda wl: _DisperisonFuncFormula1(wl, coefs)
                case "formula 2":
                    self.nFunc = lambda wl: _DisperisonFuncFormula2(wl, coefs)
                case "formula 5":
                    self.nFunc = lambda wl: _DisperisonFuncFormula5(wl, coefs)
                case _ as formula_type:
                    raise NotImplementedError(f"Unsupported formula type: {formula_type}")


# ===============================================================================
# Dispersion functions
# ===============================================================================


def _DispersionFunc4(wl: ArrayLike, coefs: NDArray[np.floating]) -> NDArray[np.floating]:
    wl = wl * 1e6
    n2 = coefs[0]

    for i in range(1, len(coefs), 4):
        if i + 2 < len(coefs):
            n2 += coefs[i] * wl ** coefs[i + 1] / (wl**2.0 - coefs[i + 2] ** coefs[i + 3])
        else:
            n2 += coefs[i] * wl ** coefs[i + 1]
    return np.sqrt(n2)


def _DispersionFunc1(wl: ArrayLike, coefs: NDArray[np.floating]) -> NDArray[np.floating]:
    wl = wl * 1e6

    n2MinOne = coefs[0]
    for i in range(1, len(coefs), 2):
        n2MinOne += coefs[i] * (wl**2.0) / (wl**2.0 - coefs[i + 1] ** 2.0)
    return np.sqrt(n2MinOne + 1.0)


def _DispersionFunc2(wl: ArrayLike, coefs: NDArray[np.floating]) -> NDArray[np.floating]:
    wl = wl * 1e6

    n2MinOne = coefs[0]
    for i in range(1, len(coefs), 2):
        n2MinOne += coefs[i] * (wl**2.0) / (wl**2.0 - coefs[i + 1])
    return np.sqrt(n2MinOne + 1.0)


def _DisperisonFuncFormula1(wl: ArrayLike, coefs: NDArray[np.floating]) -> NDArray[np.floating]:
    wl = wl * 1e6
    n2 = coefs[0] + 1.0
    for i in range(1, len(coefs), 2):
        n2 += coefs[i] * wl**2.0 / (wl**2.0 - coefs[i + 1] ** 2.0)
    return np.sqrt(n2)


def _DisperisonFuncFormula2(wl: ArrayLike, coefs: NDArray[np.floating]) -> NDArray[np.floating]:
    wl = wl * 1e6
    n2 = coefs[0] + 1.0
    for i in range(1, len(coefs), 2):
        n2 += coefs[i] * wl**2.0 / (wl**2.0 - coefs[i + 1])
    return np.sqrt(n2)


def _DisperisonFuncFormula5(wl: ArrayLike, coefs: NDArray[np.floating]) -> NDArray[np.floating]:
    wl = wl * 1e6
    n = coefs[0]
    for i in range(1, len(coefs), 2):
        n += coefs[i] * (wl ** coefs[i + 1])
    return n



"""Type stubs for the NonlinearTMM package."""

from __future__ import annotations

from NonlinearTMM._Material import Material as Material
from NonlinearTMM._SecondOrderNLTMMCython import NonlinearTMM as NonlinearTMM
from NonlinearTMM._SecondOrderNLTMMCython import NonlinearTMM as TMM
from NonlinearTMM._SecondOrderNLTMMCython import SecondOrderNLTMM as SecondOrderNLTMM
from NonlinearTMM._SecondOrderNLTMMCython import _Chi2Tensor as _Chi2Tensor
from NonlinearTMM._SecondOrderNLTMMCython import _FieldsZ as _FieldsZ
from NonlinearTMM._SecondOrderNLTMMCython import _FieldsZX as _FieldsZX
from NonlinearTMM._SecondOrderNLTMMCython import _HomogeneousWave as _HomogeneousWave
from NonlinearTMM._SecondOrderNLTMMCython import _Intensities as _Intensities
from NonlinearTMM._SecondOrderNLTMMCython import _NonlinearLayer as _NonlinearLayer
from NonlinearTMM._SecondOrderNLTMMCython import (
    _SecondOrderNLIntensities as _SecondOrderNLIntensities,
)
from NonlinearTMM._SecondOrderNLTMMCython import (
    _SweepResultNonlinearTMM as _SweepResultNonlinearTMM,
)
from NonlinearTMM._SecondOrderNLTMMCython import (
    _SweepResultSecondOrderNLTMM as _SweepResultSecondOrderNLTMM,
)
from NonlinearTMM._SecondOrderNLTMMCython import _Wave as _Wave
from NonlinearTMM._SecondOrderNLTMMCython import (
    _WaveSweepResultNonlinearTMM as _WaveSweepResultNonlinearTMM,
)
from NonlinearTMM._SecondOrderNLTMMCython import (
    _WaveSweepResultSecondOrderNLTMM as _WaveSweepResultSecondOrderNLTMM,
)

__all__ = ["Material", "TMM", "NonlinearTMM", "SecondOrderNLTMM"]



"""Type stubs for the Cython extension module _SecondOrderNLTMMCython.

This module is compiled from ``NonlinearTMM/src/SecondOrderNLTMM.pyx``.
"""

from __future__ import annotations

from typing import Any

import numpy as np
from numpy.typing import NDArray

# ---------------------------------------------------------------------------
# _Chi2Tensor
# ---------------------------------------------------------------------------

class _Chi2Tensor:
    def Update(self, **kwargs: Any) -> None: ...
    def GetChi2Tensor(self) -> NDArray[np.float64]: ...

# ---------------------------------------------------------------------------
# Material
# ---------------------------------------------------------------------------

class Material:
    chi2: _Chi2Tensor

    def __init__(self, wls: NDArray[np.float64], ns: NDArray[np.complex128]) -> None: ...
    def __call__(self, wl: float) -> complex: ...
    def GetN(self, wl: float) -> complex: ...
    def IsNonlinear(self) -> bool: ...

# ---------------------------------------------------------------------------
# _Wave
# ---------------------------------------------------------------------------

class _Wave:
    pwr: float
    overrideE0: bool
    E0: float
    w0: float
    Ly: float
    a: float
    nPointsInteg: int
    maxX: float
    dynamicMaxX: bool
    dynamicMaxXCoef: float
    dynamicMaxXAddition: float
    maxPhi: float

    def SetParams(self, **kwargs: Any) -> None: ...
    @property
    def waveType(self) -> str: ...
    @waveType.setter
    def waveType(self, value: str) -> None: ...
    @property
    def xRange(self) -> tuple[float, float]: ...
    @property
    def betas(self) -> NDArray[np.float64]: ...
    @property
    def phis(self) -> NDArray[np.float64]: ...
    @property
    def kxs(self) -> NDArray[np.float64]: ...
    @property
    def kzs(self) -> NDArray[np.complex128]: ...
    @property
    def fieldProfile(self) -> tuple[NDArray[np.float64], NDArray[np.float64]]: ...
    @property
    def expansionCoefsKx(self) -> NDArray[np.complex128]: ...
    @property
    def beamArea(self) -> float: ...

# ---------------------------------------------------------------------------
# _Intensities
# ---------------------------------------------------------------------------

class _Intensities:
    @property
    def inc(self) -> complex: ...
    @property
    def r(self) -> complex: ...
    @property
    def t(self) -> complex: ...
    @property
    def I(self) -> float: ...  # noqa: E743
    @property
    def R(self) -> float: ...
    @property
    def T(self) -> float: ...

# ---------------------------------------------------------------------------
# _SweepResultNonlinearTMM
# ---------------------------------------------------------------------------

class _SweepResultNonlinearTMM:
    @property
    def inc(self) -> NDArray[np.complex128]: ...
    @property
    def r(self) -> NDArray[np.complex128]: ...
    @property
    def t(self) -> NDArray[np.complex128]: ...
    @property
    def Ii(self) -> NDArray[np.float64]: ...
    @property
    def Ir(self) -> NDArray[np.float64]: ...
    @property
    def It(self) -> NDArray[np.float64]: ...
    @property
    def Ia(self) -> NDArray[np.float64]: ...
    @property
    def enh(self) -> NDArray[np.float64]: ...

# ---------------------------------------------------------------------------
# _WaveSweepResultNonlinearTMM
# ---------------------------------------------------------------------------

class _WaveSweepResultNonlinearTMM:
    @property
    def Pi(self) -> NDArray[np.float64]: ...
    @property
    def Pr(self) -> NDArray[np.float64]: ...
    @property
    def Pt(self) -> NDArray[np.float64]: ...
    @property
    def enh(self) -> NDArray[np.float64]: ...
    @property
    def beamArea(self) -> NDArray[np.float64]: ...

# ---------------------------------------------------------------------------
# _FieldsZ
# ---------------------------------------------------------------------------

class _FieldsZ:
    @property
    def E(self) -> NDArray[np.complex128]: ...
    @property
    def H(self) -> NDArray[np.complex128]: ...

# ---------------------------------------------------------------------------
# _FieldsZX
# ---------------------------------------------------------------------------

class _FieldsZX:
    @property
    def Ex(self) -> NDArray[np.complex128] | None: ...
    @property
    def Ey(self) -> NDArray[np.complex128] | None: ...
    @property
    def Ez(self) -> NDArray[np.complex128] | None: ...
    @property
    def Hx(self) -> NDArray[np.complex128] | None: ...
    @property
    def Hy(self) -> NDArray[np.complex128] | None: ...
    @property
    def Hz(self) -> NDArray[np.complex128] | None: ...
    @property
    def EN(self) -> NDArray[np.float64]: ...
    @property
    def HN(self) -> NDArray[np.float64]: ...

# ---------------------------------------------------------------------------
# _HomogeneousWave
# ---------------------------------------------------------------------------

class _HomogeneousWave:
    def GetMainFields(self, z: float) -> NDArray[np.complex128]: ...
    @property
    def kzF(self) -> complex: ...
    @property
    def kx(self) -> float: ...

# ---------------------------------------------------------------------------
# _NonlinearLayer
# ---------------------------------------------------------------------------

class _NonlinearLayer:
    d: float
    hw: _HomogeneousWave

    def GetIntensity(self, z: float) -> float: ...
    def GetAbsorbedIntensity(self) -> float: ...
    def GetSrcIntensity(self) -> float: ...

# ---------------------------------------------------------------------------
# NonlinearTMM
# ---------------------------------------------------------------------------

class NonlinearTMM:
    wl: float
    beta: float
    pol: str
    I0: float
    overrideE0: bool
    E0: complex
    mode: str

    def __init__(
        self,
        *,
        initStruct: bool = ...,
        parent: object = ...,
        **kwargs: Any,
    ) -> None: ...
    @property
    def materialsCache(self) -> list[Material]: ...
    @property
    def layers(self) -> list[_NonlinearLayer]: ...
    @property
    def wave(self) -> _Wave: ...
    def AddLayer(self, d: float, material: Material) -> None: ...
    def SetParams(self, **kwargs: Any) -> None: ...
    def Solve(self, **kwargs: Any) -> None: ...
    def GetIntensities(self) -> _Intensities: ...
    def Sweep(
        self,
        paramStr: str,
        values: NDArray[np.float64],
        layerNr: int = ...,
        layerZ: float = ...,
        outPwr: bool = ...,
        outAbs: bool = ...,
        outEnh: bool = ...,
    ) -> _SweepResultNonlinearTMM: ...
    def GetFields(
        self,
        zs: NDArray[np.float64],
        dir: str = ...,
    ) -> _FieldsZ: ...
    def GetFields2D(
        self,
        zs: NDArray[np.float64],
        xs: NDArray[np.float64],
        dir: str = ...,
    ) -> _FieldsZX: ...
    def GetAbsorbedIntensity(self) -> float: ...
    def GetEnhancement(self, layerNr: int, z: float = ...) -> float: ...
    def WaveGetPowerFlows(
        self,
        layerNr: int,
        x0: float = ...,
        x1: float = ...,
        z: float = ...,
    ) -> tuple[float, float]: ...
    def WaveGetEnhancement(self, layerNr: int, z: float = ...) -> float: ...
    def WaveSweep(
        self,
        paramStr: str,
        values: NDArray[np.float64],
        layerNr: int = ...,
        layerZ: float = ...,
        outPwr: bool = ...,
        outR: bool = ...,
        outT: bool = ...,
        outEnh: bool = ...,
    ) -> _WaveSweepResultNonlinearTMM: ...
    def WaveGetFields2D(
        self,
        zs: NDArray[np.float64],
        xs: NDArray[np.float64],
        dirStr: str = ...,
    ) -> _FieldsZX: ...

# ---------------------------------------------------------------------------
# _SecondOrderNLIntensities
# ---------------------------------------------------------------------------

class _SecondOrderNLIntensities:
    @property
    def P1(self) -> _Intensities: ...
    @property
    def P2(self) -> _Intensities: ...
    @property
    def Gen(self) -> _Intensities: ...

# ---------------------------------------------------------------------------
# _SweepResultSecondOrderNLTMM
# ---------------------------------------------------------------------------

class _SweepResultSecondOrderNLTMM:
    @property
    def P1(self) -> _SweepResultNonlinearTMM: ...
    @property
    def P2(self) -> _SweepResultNonlinearTMM: ...
    @property
    def Gen(self) -> _SweepResultNonlinearTMM: ...
    @property
    def wlsGen(self) -> NDArray[np.float64]: ...
    @property
    def betasGen(self) -> NDArray[np.float64]: ...

# ---------------------------------------------------------------------------
# _WaveSweepResultSecondOrderNLTMM
# ---------------------------------------------------------------------------

class _WaveSweepResultSecondOrderNLTMM:
    @property
    def P1(self) -> _WaveSweepResultNonlinearTMM: ...
    @property
    def P2(self) -> _WaveSweepResultNonlinearTMM: ...
    @property
    def Gen(self) -> _WaveSweepResultNonlinearTMM: ...
    @property
    def wlsGen(self) -> NDArray[np.float64]: ...
    @property
    def betasGen(self) -> NDArray[np.float64]: ...

# ---------------------------------------------------------------------------
# SecondOrderNLTMM
# ---------------------------------------------------------------------------

class SecondOrderNLTMM:
    deltaWlSpdc: float
    solidAngleSpdc: float
    deltaThetaSpdc: float

    def __init__(self, mode: str = ..., **kwargs: Any) -> None: ...
    @property
    def P1(self) -> NonlinearTMM: ...
    @property
    def P2(self) -> NonlinearTMM: ...
    @property
    def Gen(self) -> NonlinearTMM: ...
    def SetParams(self, **kwargs: Any) -> None: ...
    def AddLayer(self, d: float, material: Material) -> None: ...
    def Solve(self) -> None: ...
    def UpdateGenParams(self) -> None: ...
    def GetIntensities(self) -> _SecondOrderNLIntensities: ...
    def Sweep(
        self,
        paramStr: str,
        valuesP1: NDArray[np.float64],
        valuesP2: NDArray[np.float64],
        layerNr: int = ...,
        layerZ: float = ...,
        outPwr: bool = ...,
        outAbs: bool = ...,
        outEnh: bool = ...,
        outP1: bool = ...,
        outP2: bool = ...,
        outGen: bool = ...,
    ) -> _SweepResultSecondOrderNLTMM: ...
    def WaveGetPowerFlows(
        self,
        layerNr: int,
        x0: float = ...,
        x1: float = ...,
        z: float = ...,
    ) -> tuple[float, float]: ...
    def WaveSweep(
        self,
        paramStr: str,
        valuesP1: NDArray[np.float64],
        valuesP2: NDArray[np.float64],
        layerNr: int = ...,
        layerZ: float = ...,
        outPwr: bool = ...,
        outR: bool = ...,
        outT: bool = ...,
        outEnh: bool = ...,
        outP1: bool = ...,
        outP2: bool = ...,
        outGen: bool = ...,
    ) -> _WaveSweepResultSecondOrderNLTMM: ...
    def WaveGetFields2D(
        self,
        zs: NDArray[np.float64],
        xs: NDArray[np.float64],
        dirStr: str = ...,
    ) -> _FieldsZX: ...



from __future__ import annotations

import math

import matplotlib.pyplot as plt
import numpy as np

from NonlinearTMM import Material, SecondOrderNLTMM


def CalcSHG() -> None:
    # Parameters
    # ---------------------------------------------------------------------------
    wl = 1000e-9  # Pump wavelength
    pol = "s"  # Polarization
    I0 = 1.0  # Intensity of incident pump wave
    crystalD = 1000e-6  # Crystal thickness
    betas = np.linspace(0.0, 0.99, 10000)  # Sweep range for beta

    # Define materials
    # ---------------------------------------------------------------------------
    wlsCrystal = np.array([400e-9, 1100e-9])
    nsCrystal = np.array([1.54, 1.53], dtype=complex)
    prism = Material.Static(1.0)
    crystal = Material(wlsCrystal, nsCrystal)
    crystal.chi2.Update(d22=1e-12)
    dielectric = Material.Static(1.0)

    # Init SecondOrderNLTMM
    # ---------------------------------------------------------------------------
    tmm = SecondOrderNLTMM()
    tmm.P1.SetParams(wl=wl, pol=pol, beta=0.2, I0=I0)
    tmm.P2.SetParams(wl=wl, pol=pol, beta=0.2, I0=I0)
    tmm.Gen.SetParams(pol=pol)

    # Add layers
    tmm.AddLayer(math.inf, prism)
    tmm.AddLayer(crystalD, crystal)
    tmm.AddLayer(math.inf, dielectric)

    # Beta sweep
    # ---------------------------------------------------------------------------
    sr = tmm.Sweep("beta", betas, betas, outP1=True, outGen=True)

    # Crystal thickness sweep at normal incidence (beta = 0)
    # ---------------------------------------------------------------------------
    thicknesses = np.linspace(10e-6, 2000e-6, 200)
    shg_t = np.empty(len(thicknesses))
    for i, d in enumerate(thicknesses):
        tmm2 = SecondOrderNLTMM()
        tmm2.P1.SetParams(wl=wl, pol=pol, beta=0.0, I0=I0)
        tmm2.P2.SetParams(wl=wl, pol=pol, beta=0.0, I0=I0)
        tmm2.Gen.SetParams(pol=pol)
        tmm2.AddLayer(math.inf, prism)
        tmm2.AddLayer(d, crystal)
        tmm2.AddLayer(math.inf, dielectric)
        tmm2.Solve()
        intensities = tmm2.GetIntensities()
        shg_t[i] = intensities.Gen.T

    # Plot results
    # ---------------------------------------------------------------------------
    fig, axes = plt.subplots(1, 3, figsize=(9.6, 3.2))

    # Left: Schematic of the setup
    ax = axes[0]
    ax.set_xlim(-1, 5)
    ax.set_ylim(-2, 2)
    ax.set_aspect("equal")
    ax.axis("off")
    ax.set_title("Setup")

    # Draw layers
    from matplotlib.patches import Rectangle

    ax.add_patch(Rectangle((-0.5, -1.5), 1.5, 3, fc="#ddeeff", ec="k", lw=0.8))
    ax.add_patch(Rectangle((1, -1.5), 2, 3, fc="#ffe0cc", ec="k", lw=1.2))
    ax.add_patch(Rectangle((3, -1.5), 1.5, 3, fc="#ddeeff", ec="k", lw=0.8))
    ax.text(0.25, -1.8, "air", ha="center", fontsize=8)
    ax.text(2.0, -1.8, r"$\chi^{(2)}$ crystal", ha="center", fontsize=8)
    ax.text(3.75, -1.8, "air", ha="center", fontsize=8)

    # Pump arrow
    ax.annotate(
        "",
        xy=(0.9, 0.3),
        xytext=(-0.6, 0.3),
        arrowprops=dict(arrowstyle="-|>", color="C0", lw=2),
    )
    ax.text(-0.5, 0.6, r"$\omega$ pump", fontsize=7, color="C0")

    # SHG arrows (reflected + transmitted)
    ax.annotate(
        "",
        xy=(-0.6, -0.3),
        xytext=(0.9, -0.3),
        arrowprops=dict(arrowstyle="-|>", color="C3", lw=1.5, ls="--"),
    )
    ax.text(-0.5, -0.7, r"$2\omega$ R", fontsize=7, color="C3")

    ax.annotate(
        "",
        xy=(4.6, -0.3),
        xytext=(3.1, -0.3),
        arrowprops=dict(arrowstyle="-|>", color="C3", lw=2),
    )
    ax.text(3.7, -0.7, r"$2\omega$ T", fontsize=7, color="C3")

    # Middle: SHG R, T vs beta
    ax = axes[1]
    ax.plot(betas, sr.Gen.Ir, label="R")
    ax.plot(betas, sr.Gen.It, label="T")
    ax.set_xlabel(r"$\beta$")
    ax.set_ylabel(r"Intensity ($W/m^{2}$)")
    ax.set_title(r"SHG intensity vs $\beta$")
    ax.legend()

    # Right: SHG T vs crystal thickness
    ax = axes[2]
    ax.plot(thicknesses * 1e6, shg_t)
    ax.set_xlabel(r"Crystal thickness ($\mu m$)")
    ax.set_ylabel(r"SHG transmitted ($W/m^{2}$)")
    ax.set_title(r"Thickness dependence ($\beta$ = 0)")

    fig.tight_layout()
    fig.savefig("docs/images/SecondOrderNLTMM-example.png", dpi=100)
    plt.show()


if __name__ == "__main__":
    CalcSHG()



from __future__ import annotations

from . import _SecondOrderNLTMMCython
from ._Material import Material

__all__ = [
    "Material",
    "TMM",
    "NonlinearTMM",
    "SecondOrderNLTMM",
]

# Public aliases (previously in _NonlinearTMM.py and _SecondOrderNLTMM.py)
TMM = _SecondOrderNLTMMCython.NonlinearTMM
NonlinearTMM = _SecondOrderNLTMMCython.NonlinearTMM
SecondOrderNLTMM = _SecondOrderNLTMMCython.SecondOrderNLTMM

# Helper classes
_Chi2Tensor = _SecondOrderNLTMMCython._Chi2Tensor
_Wave = _SecondOrderNLTMMCython._Wave
_Intensities = _SecondOrderNLTMMCython._Intensities
_SweepResultNonlinearTMM = _SecondOrderNLTMMCython._SweepResultNonlinearTMM
_WaveSweepResultNonlinearTMM = _SecondOrderNLTMMCython._WaveSweepResultNonlinearTMM
_FieldsZ = _SecondOrderNLTMMCython._FieldsZ
_FieldsZX = _SecondOrderNLTMMCython._FieldsZX
_HomogeneousWave = _SecondOrderNLTMMCython._HomogeneousWave
_NonlinearLayer = _SecondOrderNLTMMCython._NonlinearLayer
_SecondOrderNLIntensities = _SecondOrderNLTMMCython._SecondOrderNLIntensities
_SweepResultSecondOrderNLTMM = _SecondOrderNLTMMCython._SweepResultSecondOrderNLTMM
_WaveSweepResultSecondOrderNLTMM = _SecondOrderNLTMMCython._WaveSweepResultSecondOrderNLTMM



from __future__ import annotations

from typing import Any

import numpy as np

from NonlinearTMM import _SecondOrderNLTMMCython

__all__ = ["Material"]


class Material(_SecondOrderNLTMMCython.Material):
    __doc__ = _SecondOrderNLTMMCython.Material.__doc__

    @staticmethod
    def Static(n: complex | float) -> Material:
        """Helper method to make material with constant refractive index.

        Parameters
        ----------
        n : float or complex
            Constant value for refractive index

        Returns
        -------
        Material

        Examples
        --------
        >>> mat = Material.Static(1.5)
        >>> mat.GetN(532e-9)
        1.5

        """
        wls = np.array([-1.0, 1.0])
        ns = np.array([n, n], dtype=complex)
        return Material(wls, ns)

    @staticmethod
    def FromLabPy(materialLabPy: Any) -> Material:
        """Create a Material from a LabPy Material instance.

        Parameters
        ----------
        materialLabPy : LabPy.Material
            Source material object from the LabPy test helpers.

        Returns
        -------
        Material

        """
        if materialLabPy.materialFile == "Static":
            wls = np.array([-1.0, 1.0])
            n = materialLabPy.n + 1.0j * (materialLabPy.k + materialLabPy.kAdditional)
            ns = np.array([n, n], dtype=complex)
        elif materialLabPy.isFormula:
            wls = np.ascontiguousarray(np.linspace(materialLabPy.wlRange[0], materialLabPy.wlRange[1], 500))
            ns = np.ascontiguousarray(materialLabPy(wls))
        else:
            wls = np.ascontiguousarray(materialLabPy.wlExp)
            if materialLabPy.kExp is None:
                ns = np.ascontiguousarray(materialLabPy.nExp, dtype=complex)
            else:
                ns = np.ascontiguousarray(materialLabPy.nExp + 1.0j * materialLabPy.kExp)
            ns += 1.0j * materialLabPy.kAdditional
        res = Material(wls, ns)
        res._materialLabPy = materialLabPy  # type: ignore[reportAttributeAccessIssue]
        return res



"""This module contains physical constants and functions to convert between
different units.

Attributes:
    c (float): speed of light (m/s)
    h (float): Planck's constant (J*s)
    hp (float): Reduced Planck's constant (J*s)
    qe (float): Elementary charge (C)
    eps0 (float): Vacuum permittivity (F/m)
    mu0 (float): Vacuum permeability (A/m)

"""

from __future__ import annotations

import math
from typing import Any, TypeAlias

import numpy as np
from numpy.typing import ArrayLike, NDArray

from LabPy import Core

ScalarOrArray: TypeAlias = float | NDArray[np.floating[Any]]

# ===============================================================================
# Definition of constants
# ===============================================================================


c = 299792458.0
h = 6.62606957e-34
hp = h / (2.0 * math.pi)
qe = 1.602176565e-19
eps0 = 8.854187817e-12
mu0 = 1.2566370614e-6
pi = math.pi
kB = 1.38064852e-23


# ===============================================================================
# Wavelength to ... conversion functions
# ===============================================================================


def WlToJoule(wl: ArrayLike) -> ScalarOrArray:
    """Converts wavelength (m) to photon energy (J).

    Args:
        wl (float): wavelength (m)

    Returns:
        float: energy (J)

    """
    E = h * c / wl
    return E


def WlToEv(wl: ArrayLike) -> ScalarOrArray:
    """Converts wavelength (m) to energy (eV).

    Args:
        wl (float): wavelength

    Returns:
        float: energy (eV)

    """
    ev = h * c / (wl * qe)
    return ev


def WlToFreq(wl: ArrayLike) -> ScalarOrArray:
    """Converts wavelength to frequency (Hz).

    Args:
    wl (float): wavelength (m)

    Returns:
        float: frequency (Hz)

    """
    freq = c / wl
    return freq


def WlToOmega(wl: ArrayLike) -> ScalarOrArray:
    """Converts wavelength to angular frequency (rad/s).

    Args:
        wl (float): wavelength (m)

    Returns:
        float: angular frequency (rad/s)

    """
    omega = 2.0 * math.pi * c / wl
    return omega


# ===============================================================================
# ... to wavelength conversion functions
# ===============================================================================


def JouleToWl(E: ArrayLike) -> ScalarOrArray:
    """Converts energy of a photon (J) to wavelength (m).

    Args:
        E (float): energy (J)

    Returns:
        float: wavelength (nm)

    """
    wl = h * c / E
    return wl


def EvToWl(ev: ArrayLike) -> ScalarOrArray:
    """Converts energy (eV) to wavelength (m).

    Args:
        ev (float): energy (eV)

    Returns:
        float: wavelength (m)

    """
    wl = h * c / (ev * qe)
    return wl


def FreqToWl(freq: ArrayLike) -> ScalarOrArray:
    """Converts frequency (1/s) to wavelength (m).

    Args:
        freq (float): frequency (1/s)

    Returns:
        float: wavelength (m)

    """
    wl = c / freq
    return wl


def OmegaToWl(omega: ArrayLike) -> ScalarOrArray:
    """Converts angular frequency (rad/s) to wavelength (m).

    Args:
        omega (float): angular frequency (rad/s)

    Returns:
        float: wavelength (m)

    """
    wl = 2.0 * math.pi * c / omega
    return wl


# ===============================================================================
# Other conversion functions
# ===============================================================================


def EvToJoule(ev: ArrayLike) -> ScalarOrArray:
    """Converts eV-s to joules.

    Args:
        ev (float): Energy in eV

    Returns:
        float: Energy in J

    """

    return ev * qe


def JouleToEv(E: ArrayLike) -> ScalarOrArray:
    """Converts joules to eV-s.

    Args:
        E (float): Energy in J

    Returns:
        float: Energy in eV

    """

    return E / qe


def EvToOmega(ev: ArrayLike) -> ScalarOrArray:
    """Converts eV-s to angular frequency.

    Args:
        ev (float): Energy in eV

    Returns:
        float: Photon angular frequency (rad/s)

    """
    return 2.0 * math.pi * ev * qe / h


def OmegaToEv(omega: ArrayLike) -> ScalarOrArray:
    """Converts angular frequency to eV-s.

    Args:
        omega (float): photon angular frequency (rad/s)

    Returns:
        float: Photon energy (eV)

    """
    return omega * h / (2.0 * pi * qe)


# ===============================================================================
# Distributions
# ===============================================================================


class DetlaDistribution(Core.ParamsBaseClass):
    """This class presents delta function distribution.

    Delta function distribution is zero except at x=x0 where it is 1.0.

    Kwargs:
        x0=0.0 (float): The position of delta function

    """

    def __init__(self, **kwargs: Any) -> None:
        self._params = ["x0"]
        self.x0 = 0.0

        super().__init__(**kwargs)

    def __call__(self, x: float) -> float:
        """Returns distribution value at position x.

        Args:
            x (float): The point were to evaluate the distribution

        Returns:
            float: Distribution value at position x

        """
        if x == self.x0:
            return 1.0
        else:
            return 0.0

    def GetDist(self) -> tuple[np.ndarray, np.ndarray]:
        """Function returns (np.array([x0]), np.array([1.0])).

        Returns:
            tuple_of_arrays: Distribution positions and values

        """
        return (np.array([self.x0]), np.array([1.0]))


class NormalDistribution(Core.ParamsBaseClass):
    """This class presents normal distribution.

    Kwargs:
        x0=0.0 (float): mean
        std=1.0 (float): standard deviation
        nPoints=30 (int): number of points in distribution
        rangeB=3.0 (float): beginning of sampling range is x0 - ramgeB * std
        rangeE=3.0 (float): end of sampling range is x0 + ramgeE * std

    """

    def __init__(self, **kwargs: Any) -> None:
        self._params = ["x0", "std", "nPoints", "rangeB", "rangeE"]
        self.x0 = 50e-9
        self.std = 10e-9
        self.nPoints = 30
        self.rangeB = 3.0
        self.rangeE = 3.0

        super().__init__(**kwargs)

    def __call__(self, x: ArrayLike) -> ScalarOrArray:
        """Returns distribution value at position x.

        Args:
            x (float): The point were to evaluate the distribution

        Returns:
            float: Distribution value at position x

        """
        res = 1.0 / (self.std * math.sqrt(2.0 * math.pi)) * np.exp(-((x - self.x0) ** 2.0) / (2.0 * self.std**2.0))

        return res

    def GetDist(self) -> tuple[np.ndarray, np.ndarray]:
        """Function returns distribution at nPoints sampling points.

        The sampling range is from max(1e-16, x0 - rangeB * std) to \
        x0 + rangeE * std where nPoints sampling points are taken.

        Returns:
            tuple_of_arrays: Distribution positions and values

        """
        xs = np.linspace(max(1e-12, self.x0 - self.rangeB * self.std), self.x0 + self.rangeE * self.std, self.nPoints)

        ws = self.__call__(xs)
        return xs, ws


class LogNormalDistribution(NormalDistribution):
    """This class presents lognormal distribution.

    Kwargs:
        x0=0.0 (float): mean
        std=1.0 (float): standard deviation
        nPoints=30 (int): number of points in distribution
        rangeB=0.0 (float): beginning of sampling range
        rangeE=10.0 (float): end of sampling range

    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    def __call__(self, x: ArrayLike) -> ScalarOrArray:
        """Returns distribution value at position x.

        Args:
            x (float): The point were to evaluate the distribution

        Returns:
            float: Distribution value at position x

        """
        # x02 = math.log(self.x0 ** 2.0 / math.sqrt(self.std + self.x0 ** 2.0))
        # std2 = math.sqrt(math.log(1.0 + self.std / self.x0 ** 2.0))
        x02 = np.log(self.x0)
        std2 = self.std

        # res = 1.0 / (std2 * math.sqrt(2.0 * math.pi) * x) * \
        #    np.exp(- (np.log(x) - x02) ** 2.0 / (2.0 * std2 ** 2.0))

        res = 1.0 / (std2 * math.sqrt(2.0 * math.pi)) * np.exp(-((np.log(x) - x02) ** 2.0) / (2.0 * std2**2.0))

        return res

    def GetDist(self) -> tuple[np.ndarray, np.ndarray]:
        """Function returns distribution at nPoints sampling points.

        The sampling range is from max(1e-16, x0 - rangeB * std) to \
        x0 + rangeE * std where nPoints sampling points are taken.

        Returns:
            tuple_of_arrays: Distribution positions and values

        """
        xs = np.linspace(self.rangeB, self.rangeE, self.nPoints)

        ws = self.__call__(xs)
        return xs, ws


class LogNormalDistributionLocal(Core.ParamsBaseClass):
    """This class presents lognormal distribution with usual local parameters.

    Kwargs:
        xCoef (float): x-scale linear transformation parameter
        xOffset (float): x-offset value (new_x = x + xOffset)
        mu (float): location parameter
        sigma (float): scale parameter
        nPoint (int): number of points in distribution
        rangeB (float): beginning of sampling range
        rangeE (float): end of sampling range


    """

    def __init__(self, **kwargs: Any) -> None:
        self._params = ["xCoef", "xOffset", "mu", "sigma", "nPoints", "distB", "distE"]
        self.xCoef = 1.0
        self.xOffset = 0.0
        self.mu = 0.0
        self.sigma = 1.0
        self.nPoints = 30
        self.distB = 0.0
        self.distE = 3.0

        super().__init__(**kwargs)

    def __call__(self, x: ArrayLike) -> ScalarOrArray:
        """Returns distribution value at position x.

        Args:
            x (float): The point were to evaluate the distribution

        Returns:
            float: Distribution value at position x

        """

        x = self.xCoef * (x - self.xOffset)

        res = (
            1.0
            / (self.sigma * math.sqrt(2.0 * math.pi) * x)
            * np.exp(-((np.log(x) - self.mu) ** 2.0) / (2.0 * self.sigma**2.0))
        )

        return res

    def GetDist(self) -> tuple[np.ndarray, np.ndarray]:
        """Function returns distribution at nPoints sampling points.

        The sampling range is from max(1e-16, distB) to \
        distE where nPoints sampling points are taken.

        Returns:
            tuple_of_arrays: Distribution positions and values

        """
        xs = np.linspace(max(self.xOffset + 1e-16, self.distB), self.distE, self.nPoints)

        ws = self.__call__(xs)
        return xs, ws


# ===============================================================================
# Distribution factory
# ===============================================================================

_DISTRIBUTIONS: dict[str, type] = {
    "DetlaDistribution": DetlaDistribution,
    "NormalDistribution": NormalDistribution,
    "LogNormalDistribution": LogNormalDistribution,
    "LogNormalDistributionLocal": LogNormalDistributionLocal,
}


def GetDistributionNames() -> list[str]:
    return list(_DISTRIBUTIONS)


def Distribution(
    name: str, **kwargs: Any
) -> DetlaDistribution | NormalDistribution | LogNormalDistribution | LogNormalDistributionLocal:
    if name not in _DISTRIBUTIONS:
        raise ValueError(f"Unknown distribution name: {name}")
    return _DISTRIBUTIONS[name](**kwargs)



"""This module contains low-level methods and classes for LabPy package."""

from __future__ import annotations

import logging
import math
from bisect import bisect_left, bisect_right
from collections.abc import Callable
from time import time
from typing import Any

import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy.integrate import quad

logger = logging.getLogger(__name__)

# ===============================================================================
# Methods
# ===============================================================================


def Norm(vector: NDArray[np.complexfloating]) -> NDArray[np.floating] | np.floating:
    """Calculates vector 2nd order norm

    Args:
        vector (numpy.array): input x-array

    Returns:
        np.array: vector norm

    """
    if len(vector.shape) > 1:
        if vector.shape[1] != 3:
            raise ValueError("Only vectors with length 3 supported.")
        return np.sqrt(
            abs(vector[:, 0]) ** 2.0 + abs(vector[:, 1]) ** 2.0 + abs(vector[:, 2]) ** 2.0, dtype=complex
        ).real
    else:
        if len(vector) != 3:
            raise ValueError("Only vectors with length 3 supported.")
        return np.sqrt(abs(vector[0]) ** 2.0 + abs(vector[1]) ** 2.0 + abs(vector[2]) ** 2.0, dtype=complex).real


def PickPoints(
    x: ArrayLike,
    n: int,
    fr: float = -math.inf,
    to: float = math.inf,
) -> NDArray[np.intp]:
    """Picks n points from array x, where fr <= x <= to

    Args:
        x (numpy.array): input x-array (must be sorted)
        n (int): number of points to pick
        fr (float, optional): requirement fr <= x
        to (float, optional): requirement x >= to

    Returns:
        np.array: array of picked indices

    """

    indexFr = bisect_left(x, fr)
    indexTo = bisect_right(x, to)
    pointsTotal = indexTo - indexFr

    if pointsTotal < n:
        n = pointsTotal

    if n < 1:
        return np.array([])

    res = np.zeros(n, dtype=int)

    res[0] = indexFr
    pointsSel = 1

    pointsDensity = n / pointsTotal

    for i in range(indexFr + 1, indexTo):
        shouldBeSelected = (i - indexFr + 1) * pointsDensity
        # print "cond", shouldBeSelected, pointsSel, int(shouldBeSelected + 1e-10)
        if pointsSel < int(shouldBeSelected + 1e-10):
            res[pointsSel] = i
            pointsSel += 1

    assert pointsSel == n
    return res


def complex_quadrature(
    func: Callable[[float], complex],
    a: float,
    b: float,
    **kwargs: Any,
) -> tuple[complex, tuple[Any, ...], tuple[Any, ...]]:
    """Integrates complex function func from a to b

    Args:
        func (function): complex function to integrate
        a (float): integration parameter
        b (float): integration parameter
        kwargs (dict, optional): optimal parameters for quad

    Returns:
        tuple: (complex integration value, real integration value, imaginary integral value)

    """

    def real_func(x: float) -> float:
        return np.real(func(x))

    def imag_func(x: float) -> float:
        return np.imag(func(x))

    real_integral = quad(real_func, a, b, **kwargs)
    imag_integral = quad(imag_func, a, b, **kwargs)
    return (real_integral[0] + 1j * imag_integral[0], real_integral[1:], imag_integral[1:])


def FindZeros(
    func: Callable[[complex], complex],
    dfunc: Callable[[complex], complex],
    a: complex,
    b: complex,
    zerosMaxDepth: int = 5,
    disp: bool = True,
    **kwargs: Any,
) -> list[complex]:
    """Finds all zeros of function (in range a to b) by argument principle method

    Args:
        argPrintcipleFunc (function): function of argument principle
        a (complex): integration parameter
        b (complex): integration parameter
        zerosMaxDepth (int): maximum recursion depth
        disp (bool): display debug data
        kwargs (dict, optional): optimal parameters for integration

    Returns:
        list of complex: the list of zeros

    """

    def _ZerosInBox(ax, bx):
        start = time()

        a1x, a2x = min(ax.real, bx.real), max(ax.real, bx.real)
        b1x, b2x = min(ax.imag, bx.imag), max(ax.imag, bx.imag)

        func1 = lambda x: argPrintcipleFunc(a1x + 1.0j * x)
        func2 = lambda x: argPrintcipleFunc(x + 1.0j * b2x)
        func3 = lambda x: argPrintcipleFunc(a2x + 1.0j * x)
        func4 = lambda x: argPrintcipleFunc(x + 1.0j * b1x)

        integArgs = kwargs
        integral = 1.0j * complex_quadrature(func1, b1x, b2x, **integArgs)[0]
        integral += complex_quadrature(func2, a1x, a2x, **integArgs)[0]
        integral += 1.0j * complex_quadrature(func3, b2x, b1x, **integArgs)[0]
        integral += complex_quadrature(func4, a2x, a1x, **integArgs)[0]

        res = -integral / (2.0j * math.pi)

        if disp:
            logger.debug("zeros in box %s %s %s", ax, bx, res)
            logger.debug("time for box: %s", time() - start)

        return res

    def _DoSearch(a, b, depth=0):
        zerosRaw = _ZerosInBox(a, b).real
        logger.debug("DoSearch %s %s %s zeros %s", a, b, depth, zerosRaw)
        zeros = int(round(zerosRaw))
        if zeros <= 0:
            return []

        if depth > zerosMaxDepth:
            return [(a + b) / 2.0]

        a1, a3 = min(a.real, b.real), max(a.real, b.real)
        b1, b3 = min(a.imag, b.imag), max(a.imag, b.imag)
        a2, b2 = (a1 + a3) / 2.0, (b1 + b3) / 2.0

        res = []
        res += _DoSearch(complex(a1, b1), complex(a2, b2), depth + 1)
        res += _DoSearch(complex(a2, b1), complex(a3, b2), depth + 1)
        res += _DoSearch(complex(a1, b2), complex(a2, b3), depth + 1)
        res += _DoSearch(complex(a2, b2), complex(a3, b3), depth + 1)
        return res

    argPrintcipleFunc = lambda x: dfunc(x) / func(x)
    res = _DoSearch(a, b, 0)
    return res


def ExtraInterpolate(xs: ArrayLike, ys: ArrayLike, x: float) -> float:
    if len(xs) == 1:
        return ys[0]
    elif len(xs) == 2:
        return (x * ys[0] - x * ys[1] + xs[0] * ys[1] - xs[1] * ys[0]) / (xs[0] - xs[1])
    elif len(xs) == 3:
        return (
            (
                (ys[2] * xs[1] - ys[1] * xs[2] + x * (ys[1] - ys[2])) * xs[0] ** 2
                + (-ys[2] * xs[1] ** 2 + ys[1] * xs[2] ** 2 - x**2 * (ys[1] - ys[2])) * xs[0]
                + (ys[0] * xs[2] - x * (ys[0] - ys[2])) * xs[1] ** 2
                + (-ys[0] * xs[2] ** 2 + x**2 * (ys[0] - ys[2])) * xs[1]
                - xs[2] * x * (ys[0] - ys[1]) * (x - xs[2])
            )
            / (xs[1] - xs[2])
            / (xs[0] - xs[2])
            / (xs[0] - xs[1])
        )
    else:
        raise NotImplementedError(f"len = {len(xs)}")


def RotationMatrixX(phi: float) -> NDArray[np.floating]:
    """Rotation matrix around X-axis

    Args:
        phi (float): roatation angle

    Returns:
        numpy array (3, 3): rotation matrix

    """
    res = np.array([[1.0, 0.0, 0.0], [0.0, np.cos(phi), -np.sin(phi)], [0.0, np.sin(phi), np.cos(phi)]], dtype=float)
    return res


def RotationMatrixY(phi: float) -> NDArray[np.floating]:
    """Rotation matrix around Y-axis

    Args:
        phi (float): roatation angle

    Returns:
        numpy array (3, 3): rotation matrix

    """
    res = np.array([[np.cos(phi), 0.0, np.sin(phi)], [0.0, 1.0, 0.0], [-np.sin(phi), 0.0, np.cos(phi)]], dtype=float)
    return res


def RotationMatrixZ(phi: float) -> NDArray[np.floating]:
    """Rotation matrix around Z-axis

    Args:
        phi (float): roatation angle

    Returns:
        numpy array (3, 3): rotation matrix

    """
    res = np.array([[np.cos(phi), -np.sin(phi), 0.0], [np.sin(phi), np.cos(phi), 0.0], [0.0, 0.0, 1.0]], dtype=float)
    return res


def CartesianProduct(
    arrays: list[ArrayLike],
    out: NDArray | None = None,
) -> NDArray:
    """
    Generate a cartesian product of input arrays.

    Parameters
    ----------
    arrays : list of array-like
        1-D arrays to form the cartesian product of.
    out : ndarray
        Array to place the cartesian product in.

    Returns
    -------
    out : ndarray
        2-D array of shape (M, len(arrays)) containing cartesian products
        formed of input arrays.

    Examples
    --------
    >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
           [1, 4, 7],
           [1, 5, 6],
           [1, 5, 7],
           [2, 4, 6],
           [2, 4, 7],
           [2, 5, 6],
           [2, 5, 7],
           [3, 4, 6],
           [3, 4, 7],
           [3, 5, 6],
           [3, 5, 7]])

    """

    arrays = [np.asarray(x) for x in arrays]
    dtype = arrays[0].dtype

    n = np.prod([x.size for x in arrays])
    if out is None:
        out = np.zeros([n, len(arrays)], dtype=dtype)

    m = n // arrays[0].size
    out[:, 0] = np.repeat(arrays[0], m)
    if arrays[1:]:
        CartesianProduct(arrays[1:], out=out[0:m, 1:])
        for j in range(1, arrays[0].size):
            out[j * m : (j + 1) * m, 1:] = out[0:m, 1:]
    return out


# ===============================================================================
# Classes
# ===============================================================================


class ParamsBaseClass:
    """Inheritance of this class will provide SetParams and GetParams methods.

    Args:
        kwargs: optional arguments, currently unused

    """

    def __init__(self, params: list[str] | None = None, **kwargs: Any) -> None:
        if params is not None:
            self._params = params

        if not hasattr(self, "friendlyNames"):
            self.friendlyNames = self._params[:]

        self.SetParams(**kwargs)

    def SetParams(self, **kwargs: Any) -> None:
        """Sets parameters from kwargs dictionary.

        Args:
            kwargs (dict): dictionary of parameters

        Raises:
            ValueError: if parameter name is not in _params list

        """

        for k, v in kwargs.items():
            if k not in self._params:
                raise ValueError(f"Unknown kwargs: {k}")
            setattr(self, k, v)

    def GetParams(self, asList: bool = False) -> dict[str, Any] | list[tuple[str, str, Any]]:
        """Returns values of parameters in _params.

        Args:
            asList (bool, optional = False): Whether return params as list or as dict.

        Returns:
            dict: dictionary of parameters and values

        """
        if asList:
            res = [
                (param, friendlyName, getattr(self, param))
                for param, friendlyName in zip(self._params, self.friendlyNames, strict=True)
            ]

        else:
            res = {param: getattr(self, param) for param in self._params}
        return res



import glob
import platform
import sys

import eigency
import numpy as np
from Cython.Build import cythonize
from setuptools import setup
from setuptools.extension import Extension

extra_compile_args = []
extra_link_args = []

if sys.platform == "linux":
    extra_compile_args.extend(["-std=c++17", "-fopenmp"])
    if platform.machine() in ("x86_64", "AMD64", "i686", "i386"):
        extra_compile_args.append("-msse3")
    extra_link_args.extend(["-lgomp", "-fopenmp"])
elif sys.platform == "darwin":
    extra_compile_args.append("-std=c++17")
elif sys.platform == "win32":
    extra_compile_args.extend(["/std:c++17", "/openmp"])

sources = ["NonlinearTMM/src/SecondOrderNLTMM.pyx"] + glob.glob("NonlinearTMM/src/cpp/*.cpp")

extensions = cythonize(
    [
        Extension(
            "NonlinearTMM._SecondOrderNLTMMCython",
            sources=sources,
            include_dirs=[
                np.get_include(),
                "NonlinearTMM/src/cpp",
                "third_party/eigen",
            ]
            + eigency.get_includes(include_eigen=False),
            language="c++",
            extra_compile_args=extra_compile_args,
            extra_link_args=extra_link_args,
        ),
    ],
)

setup(ext_modules=extensions)



from __future__ import annotations

import math
from typing import Any

import numpy as np
from numpy.typing import NDArray

from LabPy import Constants, Core
from LabPy.Constants import OmegaToWl, WlToOmega

__all__ = ["Chi2Tensor", "SecondOrderNLTmm"]

X = 0
Y = 1
Z = 2

# ===============================================================================
# Chi2Tensor
# ===============================================================================


class Chi2Tensor:
    def __init__(self, dtype: type = complex, distinctFields: bool = True, **kwargs: Any) -> None:
        self._dIndices = [(0, 0), (1, 1), (2, 2), (1, 2), (0, 2), (0, 1)]
        self._chi2 = np.zeros((3, 3, 3), dtype=dtype)
        self._chi2Rotated = None
        self._chi2RotatedChanged = True

        self.distinctFields = distinctFields
        self.phiX = 0.0
        self.phiY = 0.0
        self.phiZ = 0.0

        self.Update(**kwargs)

    def Update(self, **kwargs: Any) -> None:
        self._chi2RotatedChanged = True
        self.phiX = float(kwargs.pop("phiX", self.phiX))
        self.phiY = float(kwargs.pop("phiY", self.phiY))
        self.phiZ = float(kwargs.pop("phiZ", self.phiZ))

        # d-values
        for k in list(kwargs):
            if (not k.startswith("d")) or (len(k) != 3):
                continue
            i1, (i2, i3) = int(k[1]) - 1, self._dIndices[int(k[2]) - 1]
            self._chi2[i1, i2, i3] = 2.0 * kwargs.pop(k)
            self._chi2[i1, i3, i2] = self._chi2[i1, i2, i3]

        # chi2-values
        for k in list(kwargs):
            if (not k.startswith("chi")) or (len(k) != 6):
                continue
            i1, i2, i3 = int(k[3]) - 1, int(k[4]) - 1, int(k[5]) - 1
            self._chi2[i1, i2, i3] = kwargs.pop(k)

        # Errors
        for k in list(kwargs):
            raise ValueError(f"Unknown kwarg {k}")

    def GetChi2(self) -> NDArray:
        if not self._chi2RotatedChanged:
            return self._chi2Rotated

        # Compute chi2Rotated
        self._chi2Rotated = Chi2Tensor._RotateTensorO3(self._chi2, self.phiX, self.phiY, self.phiZ)
        self._chi2RotatedChanged = False
        return self._chi2Rotated

    def GetD(self) -> NDArray:
        res = np.zeros((3, 6), dtype=self._chi2.dtype)
        chi2 = self.GetChi2()

        for i in range(3):
            for j in range(6):
                a, b = self._dIndices[j]
                if abs(chi2[i, a, b] - chi2[i, b, a]) > 1e-20:
                    raise ValueError(f"Not equal elements {(i, a, b)}, {(i, b, a)}")

                res[i, j] = chi2[i, a, b]
        return res

    def GetNonlinearPolarization(self, E1: NDArray, E2: NDArray) -> NDArray:
        chi2 = self.GetChi2()
        fieldTensor = np.outer(E1, E2)
        res = Constants.eps0 * np.tensordot(chi2, fieldTensor).flatten()

        if self.distinctFields:
            res *= 2.0
        return res

    def GetChi2Eff(self, E1: NDArray, E2: NDArray, E3: NDArray) -> complex:
        E1Norm = (E1 / np.linalg.norm(E1)).flatten()
        E2Norm = (E2 / np.linalg.norm(E2)).flatten()
        E3Norm = (E3 / np.linalg.norm(E3)).flatten()
        p = self.GetNonlinearPolarization(E1Norm, E2Norm)

        res = np.dot(E3Norm, p / Constants.eps0)
        if self.distinctFields:
            res *= 2.0
        return res

    def GetDEff(self, E1: NDArray, E2: NDArray, E3: NDArray) -> complex:
        chi2Eff = self.GetChi2Eff(E1, E2, E3)
        res = chi2Eff / 2.0
        return res

    @staticmethod
    def _RotateTensorO3(
        tensor: NDArray,
        phiX: float = 0.0,
        phiY: float = 0.0,
        phiZ: float = 0.0,
    ) -> NDArray:

        def ApplyRotationMatrix(inp: NDArray, R: NDArray) -> NDArray:
            res = np.zeros_like(inp)
            for i1 in range(3):
                for i2 in range(3):
                    for i3 in range(3):
                        for j1 in range(3):
                            for j2 in range(3):
                                for j3 in range(3):
                                    res[i1, i2, i3] += R[i1, j1] * R[i2, j2] * R[i3, j3] * inp[j1, j2, j3]

            return res

        Rx = Core.RotationMatrixX(phiX)
        Ry = Core.RotationMatrixY(phiY)
        Rz = Core.RotationMatrixZ(phiZ)

        res = ApplyRotationMatrix(tensor, Rx)
        res = ApplyRotationMatrix(res, Ry)
        res = ApplyRotationMatrix(res, Rz)

        return res


# ===============================================================================
# _HomogeneousWave
# ===============================================================================


class _HomogeneousWave:
    def __init__(self, layer: _NonlinearLayer) -> None:
        self.layer = layer

    def Solve(self, kx: complex) -> None:
        layer = self.layer

        self.kx = kx
        self.k = self.layer.k0 * self.layer.n(self.layer.wl)
        self.kzF = np.sqrt(self.k**2.0 - self.kx**2.0, dtype=complex)
        self.kzB = -self.kzF
        self.kz = np.array([[self.kzF, self.kzB]]).T

        if self.kzF.imag < 0.0:
            raise ValueError(f"kzF imaginary part negative {self.kzF}")

        # Propagation matrix

        if layer.d == math.inf:
            self.propMatrix = np.identity(2, dtype=complex)
        else:
            phase = np.exp(1.0j * self.kzF * layer.d)
            self.propMatrix = np.array([[phase, 0.0], [0.0, 1.0 / phase]], dtype=complex)

    def GetMainFields(self, zs: NDArray) -> NDArray:
        U0 = np.array([[self.layer.fieldAmps[0, 0], self.layer.fieldAmps[1, 0]]]).T
        U = U0 * np.exp(1.0j * self.kz * zs)
        return U


# ===============================================================================
# _InhomogeneosWave
# ===============================================================================


class _InhomogeneosWave:
    def __init__(self, layer: _NonlinearLayer) -> None:
        self.layer = layer

    def Solve(self, kx: complex, kpS: tuple | None) -> None:
        layer = self.layer
        kSzF, pF, pB = [0.0, [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] if kpS is None else kpS

        self.kx = kx
        self.kSzF = kSzF
        self.kSzB = -kSzF
        self.kSz = np.array([[self.kSzF, self.kSzB]]).T
        self.kS = np.sqrt(self.kx**2.0 + self.kSzF**2.0, dtype=complex)

        self.pxF, self.pyF, self.pzF = [0.0, 0.0, 0.0] if pF is None else pF
        self.pxB, self.pyB, self.pzB = [0.0, 0.0, 0.0] if pB is None else pB
        self.px = np.array([[self.pxF, self.pxB]]).T
        self.py = np.array([[self.pyF, self.pyB]]).T
        self.pz = np.array([[self.pzF, self.pzB]]).T

        if self.kS.imag < 0.0:
            # print ValueError("kS imaginary part negative %s" % (self.kS))
            self.kS *= -1.0

        # Check if layer is nonlinear

        if layer.pol == "s":
            self.isNonlinearF = abs(self.pyF) > 1e-200
            self.isNonlinearB = abs(self.pyB) > 1e-200
        elif layer.pol == "p":
            self.isNonlinearF = abs(self.pxF) > 1e-200 or abs(self.pzF) > 1e-200
            self.isNonlinearB = abs(self.pxB) > 1e-200 or abs(self.pzB) > 1e-200
        else:
            raise NotImplementedError("Unknown polarization.")
        self.isNonlinear = self.isNonlinearF or self.isNonlinearB

        # Calc B-s

        if not self.isNonlinear:
            self.ByF = 0.0
            self.ByB = 0.0
        else:
            if layer.pol == "s":
                BFunc = lambda py: -(py / Constants.eps0) * layer.k0**2.0 / (layer.k**2.0 - self.kS**2.0)
                self.ByF = BFunc(self.pyF)
                self.ByB = BFunc(self.pyB) * np.exp(1.0j * self.kSzF * layer.d)
            elif layer.pol == "p":
                BFunc = lambda px, pz, kSz: -layer.omega / (layer.k**2.0 - self.kS**2.0) * (kSz * px - self.kx * pz)
                self.ByF = BFunc(self.pxF, self.pzF, self.kSzF)
                self.ByB = BFunc(self.pxB, self.pzB, -self.kSzF) * np.exp(1.0j * self.kSzF * layer.d)
            else:
                raise NotImplementedError("Unknown polarization.")
        self.By = np.array([[self.ByF, self.ByB]]).T

        # Propagation Matrix NL

        if layer.d == math.inf:
            self.propMatrixNL = np.array([[0.0, 0.0]], dtype=complex).T
            if self.isNonlinear:
                raise NotImplementedError("First and last medium must be linear.")
        else:
            if self.isNonlinear:
                self.propMatrixNL = self.By * (np.exp(1.0j * self.kSz * layer.d) - np.exp(1.0j * layer.hw.kz * layer.d))
            else:
                self.propMatrixNL = np.array([[0.0, 0.0]], dtype=complex).T

    def GetMainFields(self, zs: NDArray) -> NDArray:
        U = self.By * (np.exp(1.0j * self.kSz * zs) - np.exp(1.0j * self.layer.hw.kz * zs))
        return U


# ===============================================================================
# _NonlinearLayer
# ===============================================================================


class _NonlinearLayer(Core.ParamsBaseClass):
    def __init__(self, d: float, n: object, **kwargs: Any) -> None:
        self._params = ["d", "n", "kpS", "kpA"]
        self.d = d
        self.n = n
        self.kpS = None
        self.kpA = None

        self.hw = _HomogeneousWave(self)
        self.iws = _InhomogeneosWave(self)
        self.iwa = _InhomogeneosWave(self)

        super().__init__(**kwargs)

    def Solve(self, wl: float, beta: float, pol: str, lastLayer: bool = False) -> None:
        self.wl = wl
        self.beta = beta
        self.pol = pol
        self.omega = Constants.WlToOmega(wl)
        self.lastLayer = lastLayer

        # Find wavevectors

        self.eps = self.n(wl) ** 2.0
        self.k0 = 2.0 * np.pi / wl
        self.k = self.k0 * self.n(self.wl)
        kx = beta * self.k0

        # Solve waves

        self.hw.Solve(kx)
        self.iws.Solve(kx, self.kpS)
        self.iwa.Solve(kx, self.kpA)

        # Propagation matrix

        self.propMatrix = self.hw.propMatrix
        self.propMatrixNL = self.iws.propMatrixNL + self.iwa.propMatrixNL

    def GetMainFields(self, zs: NDArray) -> NDArray:
        return self.hw.GetMainFields(zs) + self.iws.GetMainFields(zs) + self.iwa.GetMainFields(zs)

    def GetFields(self, zs: NDArray, outBackAndForward: bool = False) -> tuple[NDArray, NDArray]:
        E = np.zeros((2, len(zs), 3), dtype=complex)
        H = np.zeros((2, len(zs), 3), dtype=complex)

        expKSz = np.exp(1.0j * self.iws.kSz * zs)
        expKAz = np.exp(1.0j * self.iwa.kSz * zs)
        if self.pol == "p":
            c = -1.0 / (self.omega * self.eps * Constants.eps0)
            Hy = self.GetMainFields(zs)

            H[:, :, Y] = Hy
            E[:, :, X] = c * (
                -self.hw.kz * Hy
                + self.iws.By * (self.hw.kz - self.iws.kSz) * expKSz
                + self.omega * self.iws.px * expKSz
                + self.iwa.By * (self.hw.kz - self.iwa.kSz) * expKAz
                + self.omega * self.iwa.px * expKAz
            )
            E[:, :, Z] = c * (self.hw.kx * Hy + self.omega * self.iws.pz * expKSz + self.omega * self.iwa.pz * expKAz)
        elif self.pol == "s":
            c = 1.0 / (self.omega * Constants.mu0)
            Ey = self.GetMainFields(zs)

            E[:, :, Y] = Ey
            H[:, :, X] = c * (
                -self.hw.kz * Ey
                + self.iws.By * (self.hw.kz - self.iws.kSz) * expKSz
                + self.iwa.By * (self.hw.kz - self.iwa.kSz) * expKAz
            )
            H[:, :, Z] = c * (self.hw.kx * Ey)
        else:
            raise NotImplementedError("Unknown polarization.")

        if outBackAndForward:
            return E, H
        else:
            return (E[0, :, :] + E[1, :, :]), (H[0, :, :] + H[1, :, :])

    def GetAmplitudes(self, z: float = 0.0) -> tuple[tuple[NDArray, NDArray], tuple[NDArray, NDArray]]:
        (E0FTemp, E0BTemp), (H0FTemp, H0BTemp) = self.GetFields(np.array([z]), outBackAndForward=True)

        E0F, E0B = E0FTemp[0, :], E0BTemp[0, :]
        H0F, H0B = H0FTemp[0, :], H0BTemp[0, :]
        return (E0F, E0B), (H0F, H0B)

    def GetPowerFlow(self, z: float) -> float:
        (E0F, E0B), (H0F, H0B) = self.GetAmplitudes(z=z)
        E, H = E0F + E0B, H0F + H0B
        Sz = 0.5 * (E[0] * np.conj(H[1]) - E[1] * np.conj(H[0])).real
        return Sz

    def GetAbsorbedPower(self) -> float:
        if self.eps.imag == 0.0:
            # If layer is not absorbing
            return 0.0

        (E0F, E0B), (_, __) = self.GetAmplitudes()
        kz = self.hw.kz[0, 0]
        intValue1 = -1.0j * (np.exp(1.0j * (kz - np.conj(kz)) * self.d) - 1.0) / (kz - np.conj(kz))
        intValue2 = -1.0j * (np.exp(1.0j * (kz + np.conj(kz)) * self.d) - 1.0) / (kz + np.conj(kz))
        intValue3 = 1.0j * (np.exp(-1.0j * (kz + np.conj(kz)) * self.d) - 1.0) / (kz + np.conj(kz))
        intValue4 = 1.0j * (np.exp(-1.0j * (kz - np.conj(kz)) * self.d) - 1.0) / (kz - np.conj(kz))

        intValue = (
            np.dot(E0F, np.conj(E0F)) * intValue1
            + np.dot(E0F, np.conj(E0B)) * intValue2
            + np.dot(E0B, np.conj(E0F)) * intValue3
            + np.dot(E0B, np.conj(E0B)) * intValue4
        )

        res = 0.5 * Constants.eps0 * self.eps.imag * self.omega * intValue
        return res.real

    def GetSrcPower(self) -> float:
        absorbed = self.GetAbsorbedPower()
        if self.d == math.inf:
            deltaS = 0.0
        else:
            deltaS = self.GetPowerFlow(self.d) - self.GetPowerFlow(0.0)

        res = absorbed + deltaS
        return res


# ===============================================================================
# _NonlinearTmm
# ===============================================================================


class _NonlinearTmm(Core.ParamsBaseClass):
    MODE_INCIDENT = "incident"
    MODE_NONLINEAR = "nonlinear"

    def __init__(self, **kwargs: Any) -> None:
        self._params = ["wl", "beta", "pol", "I0", "overrideE0", "mode"]
        self.I0 = 1e10
        self.overrideE0 = None  # E0 specified in vacuum, not in first layer
        self.layers: list[_NonlinearLayer] = []

        super().__init__(**kwargs)

    def AddLayer(self, *args: Any, **kwargs: Any) -> None:
        self.layers.append(_NonlinearLayer(*args, **kwargs))

    def Solve(self, **kwargs: Any) -> tuple:
        self.SetParams(**kwargs)

        # Solve all layers

        for i, layer in enumerate(self.layers):
            last = i == len(self.layers) - 1
            layer.Solve(self.wl, self.beta, self.pol, last)

        # Solve system matrix

        mSysL, mSysNL = self.__SystemMatrix(len(self.layers))

        # Calc reflection and transmission

        l0 = self.layers[0]
        lL = self.layers[-1]
        omega = Constants.WlToOmega(self.wl)
        if self.mode == self.MODE_INCIDENT:
            if self.overrideE0 is None:
                cosTh0 = l0.hw.kz[0, 0].real / l0.k.real
                if self.pol == "p":
                    inc = np.sqrt(
                        abs(2.0 * omega * Constants.eps0 / l0.hw.kzF.real * l0.eps.real * self.I0 * cosTh0),
                        dtype=complex,
                    )
                elif self.pol == "s":
                    inc = np.sqrt(abs(2.0 * omega * Constants.mu0 / l0.hw.kzF.real * self.I0 * cosTh0), dtype=complex)
                else:
                    raise NotImplementedError("Unknown polarization.")
            else:
                if self.pol == "p":
                    inc = self.overrideE0 * np.sqrt(l0.n(self.wl).real) * Constants.eps0 * Constants.c
                elif self.pol == "s":
                    inc = self.overrideE0 / np.sqrt(l0.n(self.wl).real)
                else:
                    raise NotImplementedError("Unknown polarization.")

            r = -inc * mSysL[1, 0] / mSysL[1, 1]
            t = inc * np.linalg.det(mSysL) / mSysL[1, 1]
        elif self.mode == self.MODE_NONLINEAR:
            inc = 0.0
            r = -mSysNL[1, 0] / mSysL[1, 1]
            t = mSysNL[0, 0] - mSysL[0, 1] / mSysL[1, 1] * mSysNL[1, 0]
        else:
            raise NotImplementedError("Unknown mode.")

        # Calc power flows

        if self.pol == "p":
            c = 1.0 / (2.0 * omega * Constants.eps0)
            I = c * l0.hw.kzF.real / l0.eps.real * abs(inc) ** 2.0
            R = c * l0.hw.kzF.real / l0.eps.real * abs(r) ** 2.0
            T = c * lL.hw.kzF.real / lL.eps.real * abs(t) ** 2.0
        elif self.pol == "s":
            c = 1.0 / (2.0 * omega * Constants.mu0)
            I = c * l0.hw.kzF.real * abs(inc) ** 2.0
            R = c * l0.hw.kzF.real * abs(r) ** 2.0
            T = c * lL.hw.kzF.real * abs(t) ** 2.0
        else:
            raise NotImplementedError("Unknown polarization.")

        # Calc field coefs in all layers

        self.layers[0].fieldAmps = np.array([[inc, r]], dtype=complex).T
        for i in range(1, len(self.layers)):
            mL, mNL = self.__SystemMatrix(i + 1)
            self.layers[i].fieldAmps = np.dot(mL, self.layers[0].fieldAmps) + mNL

        # Absorbed energy
        A = self.GetAbsorbedPower()

        return inc, r, t, I, R, T, A

    def __TransferMatrix(self, layerNr: int) -> tuple[NDArray, NDArray]:

        def TransferMatrixNL(w1: _InhomogeneosWave, w2: _InhomogeneosWave) -> NDArray:
            exp1F = np.exp(1.0j * w1.kSzF * l1.d) if l1.d != math.inf else 0.0
            exp1B = np.exp(-1.0j * w1.kSzF * l1.d) if l1.d != math.inf else 0.0
            # exp1B = 1.0 / exp1F if l1.d != float("inf") else 0.0

            f1 = (w1.kSzF - l1.hw.kzF) * (w1.ByF * exp1F - w1.ByB * exp1B)
            f2 = (w2.kSzF - l2.hw.kzF) * (w2.ByB - w2.ByF)

            if self.pol == "p":
                f3 = -l1.omega * l2.eps * (w1.pxB * exp1B + w1.pxF * exp1F) + l1.omega * l1.eps * (w2.pxB + w2.pxF)
                cNL = 1.0 / (2.0 * l1.eps * l2.hw.kzF) * (l2.eps * f1 + l1.eps * f2 + f3)
            elif self.pol == "s":
                cNL = 1.0 / (2.0 * l2.hw.kzF) * (f1 + f2)
            else:
                raise NotImplementedError("Unknown polarization.")

            transferMatrixNL = np.array([[cNL, -cNL]], dtype=complex).T
            return transferMatrixNL

        l1, l2 = self.layers[layerNr], self.layers[layerNr + 1]

        # Polarization specific

        if self.pol == "p":
            a = (l2.eps * l1.hw.kzF) / (l1.eps * l2.hw.kzF)
        elif self.pol == "s":
            a = l1.hw.kzF / l2.hw.kzF
        else:
            raise NotImplementedError("Unknown polarization.")

        # Transfer matrices
        transferMatrix = 0.5 * np.array([[1.0 + a, 1.0 - a], [1.0 - a, 1.0 + a]], dtype=complex)
        transferMatrixNL = TransferMatrixNL(l1.iws, l2.iws) + TransferMatrixNL(l1.iwa, l2.iwa)

        return transferMatrix, transferMatrixNL

    def __SystemMatrix(self, nLayers: int) -> tuple[NDArray, NDArray]:
        # Independent of polarization

        mSysL = np.identity(2, dtype=complex)
        mSysNL = np.array([[0.0, 0.0]], dtype=complex).T

        for i in range(nLayers - 1):
            layer = self.layers[i]
            transferL, transferNL = self.__TransferMatrix(i)

            mSysL = np.dot(layer.propMatrix, mSysL)
            mSysL = np.dot(transferL, mSysL)

            mSysNL = np.dot(layer.propMatrix, mSysNL) + layer.propMatrixNL
            mSysNL = np.dot(transferL, mSysNL) + transferNL

        return mSysL, mSysNL

    def GetFields(self, zs: NDArray, outBackAndForward: bool = False) -> tuple[NDArray, NDArray]:
        if not outBackAndForward:
            resE = np.zeros((len(zs), 3), dtype=complex)
            resH = np.zeros((len(zs), 3), dtype=complex)
        else:
            resE = np.zeros((2, len(zs), 3), dtype=complex)
            resH = np.zeros((2, len(zs), 3), dtype=complex)
        layerDsIndexes = self.__GetLayerIndexesAndDistances(zs)

        for layerId, layerDist, bIndex, eIndex in layerDsIndexes:
            if bIndex == eIndex:
                continue

            E, H = self.layers[layerId].GetFields(zs[bIndex:eIndex] - layerDist, outBackAndForward=outBackAndForward)
            if not outBackAndForward:
                resE[bIndex:eIndex, :] = E
                resH[bIndex:eIndex, :] = H
            else:
                resE[:, bIndex:eIndex, :] = E
                resH[:, bIndex:eIndex, :] = H

        return resE, resH

    def GetFields2D(self, zs: NDArray, xs: NDArray, outBackAndForward: bool = False) -> tuple[NDArray, NDArray]:
        E1D, H1D = self.GetFields(zs, outBackAndForward=outBackAndForward)
        phaseX = np.exp(1.0j * 2.0 * np.pi / self.wl * self.beta * xs)

        if not outBackAndForward:
            resE = np.zeros((len(zs), len(xs), 3), dtype=complex)
            resH = np.zeros((len(zs), len(xs), 3), dtype=complex)

            for i in range(3):
                resE[:, :, i] = np.outer(E1D[:, i], phaseX)
                resH[:, :, i] = np.outer(H1D[:, i], phaseX)

        else:
            resE = np.zeros((2, len(zs), len(xs), 3), dtype=complex)
            resH = np.zeros((2, len(zs), len(xs), 3), dtype=complex)

            for j in range(2):
                for i in range(3):
                    resE[j, :, :, i] = np.outer(E1D[j, :, i], phaseX)
                    resH[j, :, :, i] = np.outer(H1D[j, :, i], phaseX)

        return resE, resH

    def GetAbsorbedPower(self) -> float:
        return sum(layer.GetAbsorbedPower() for layer in self.layers)

    def __GetLayerIndexesAndDistances(self, zs: NDArray) -> list[tuple[int, float, int, int]]:
        layerDsIndexes = []
        curLayer, curDist, layerDist, beginIndex = 0, 0.0, 0.0, 0
        for i in range(len(zs) + 1):
            if (i == len(zs)) or (zs[i] >= curDist):
                layerDsIndexes.append((curLayer, layerDist, beginIndex, i))
                if i == len(zs):
                    break
                layerDist += self.layers[curLayer].d if curLayer > 0 else 0.0
                curLayer += 1
                curDist += self.layers[curLayer].d
                beginIndex = i
        return layerDsIndexes


# ===============================================================================
# SecondOrderNLTmm
# ===============================================================================


class SecondOrderNLTmm(Core.ParamsBaseClass):
    SFG = "sfg"
    DFG = "dfg"

    def __init__(self, **kwargs: Any) -> None:
        self._params = [
            "I0P1",
            "wlP1",
            "betaP1",
            "polP1",
            "overrideE0P1",
            "I0P2",
            "wlP2",
            "betaP2",
            "polP2",
            "overrideE0P2",
            "polGen",
            "process",
        ]
        self.layers: list[tuple] = []
        self.I0P1 = 1e10
        self.I0P2 = 1e10
        self.overrideE0P1 = None
        self.overrideE0P2 = None
        self.process = self.SFG
        super().__init__(**kwargs)

    def AddLayer(self, d: float, n: object, chi2: Chi2Tensor | None = None) -> None:
        if chi2 is None:
            chi2 = Chi2Tensor()
        self.layers.append((d, n, chi2))

    def Solve(self, **kwargs: Any) -> tuple:
        self.SetParams(**kwargs)

        self._SolveFundamentalFields()

        if self.process == self.SFG:
            self.wlGen = OmegaToWl(WlToOmega(self.wlP1) + WlToOmega(self.wlP2))
            self.betaGen = self.wlGen * (self.betaP1 / self.wlP1 + self.betaP2 / self.wlP2)
            kSzFunc = lambda kzFP1, kzFP2: kzFP1 + kzFP2
            kAzFunc = lambda kzFP1, kzFP2: kzFP1 - kzFP2
            pFunc = lambda chi2, E1, E2: chi2.GetNonlinearPolarization(E1, E2)
        elif self.process == self.DFG:
            self.wlGen = OmegaToWl(WlToOmega(self.wlP1) - WlToOmega(self.wlP2))
            self.betaGen = self.wlGen * (self.betaP1 / self.wlP1 - self.betaP2 / self.wlP2)
            kSzFunc = lambda kzFP1, kzFP2: kzFP1 - np.conj(kzFP2)
            kAzFunc = lambda kzFP1, kzFP2: kzFP1 + np.conj(kzFP2)
            pFunc = lambda chi2, E1, E2: chi2.GetNonlinearPolarization(E1, np.conj(E2))
        else:
            raise ValueError(f"Unknown process: {self.process}")

        self.tmmGen = _NonlinearTmm(wl=self.wlGen, pol=self.polGen, mode=_NonlinearTmm.MODE_NONLINEAR)

        for i, (d, n, chi2) in enumerate(self.layers):
            (E0P1F, E0P1B), _ = self.tmmP1.layers[i].GetAmplitudes()
            (E0P2F, E0P2B), _ = self.tmmP2.layers[i].GetAmplitudes()
            kzFP1, kzFP2 = self.tmmP1.layers[i].hw.kzF, self.tmmP2.layers[i].hw.kzF

            kSz, kAz = kSzFunc(kzFP1, kzFP2), kAzFunc(kzFP1, kzFP2)
            pSF, pSB = pFunc(chi2, E0P1F, E0P2F), pFunc(chi2, E0P1B, E0P2B)
            pAF, pAB = pFunc(chi2, E0P1F, E0P2B), pFunc(chi2, E0P1B, E0P2F)

            if kSz.imag < 0.0:
                # print "reversed kSz"
                kSz, pSF, pSB = -kSz, pSB, pSF

            if kAz.imag < 0.0:
                # print "reversed kAz"
                kAz, pAF, pAB = -kAz, pAB, pAF

            self.tmmGen.AddLayer(d, n, kpS=(kSz, pSF, pSB), kpA=(kAz, pAF, pAB))

        self.resGen = self.tmmGen.Solve(beta=self.betaGen)

        return self.resP1, self.resP2, self.resGen

    def Sweep(
        self,
        paramNames: list[str],
        paramValues: list[NDArray],
        enhpos: tuple[int, float] | None = None,
    ) -> dict[str, NDArray]:
        keysP1 = ["iP1", "rP1", "tP1", "IP1", "RP1", "TP1", "AP1"]
        keysP2 = ["iP2", "rP2", "tP2", "IP2", "RP2", "TP2", "AP2"]
        keysGen = ["iGen", "rGen", "tGen", "IGen", "RGen", "TGen", "AGen"]
        res = {}

        res["betasGen"] = np.zeros(len(paramValues[0]))

        for k in keysP1 + keysP2 + keysGen:
            res[k] = np.zeros((len(paramValues[0])), dtype=complex)

        if enhpos is not None:
            res["enhP1"] = np.zeros(len(paramValues[0]))
            res["enhP2"] = np.zeros(len(paramValues[0]))

        for i in range(len(paramValues[0])):
            for name, values in zip(paramNames, paramValues):
                self.SetParams(**{name: values[i]})

            rrP1, rrP2, rrGen = self.Solve()
            res["betasGen"][i] = self.tmmGen.beta.real

            for key, val in zip(keysP1, rrP1):
                res[key][i] = val
            for key, val in zip(keysP2, rrP2):
                res[key][i] = val
            for key, val in zip(keysGen, rrGen):
                res[key][i] = val

            if enhpos is not None:
                enhLayer, enhDist = enhpos
                EP1, _ = self.tmmP1.layers[enhLayer].GetFields(np.array([enhDist]))
                (E0P1, _), __ = self.tmmP1.layers[0].GetFields(np.array([0.0]), outBackAndForward=True)
                enhP1 = np.linalg.norm(EP1[0, :]) / (
                    np.linalg.norm(E0P1[0, :]) * np.sqrt(self.tmmP1.layers[0].n(self.wlP1).real)
                )

                EP2, _ = self.tmmP2.layers[enhLayer].GetFields(np.array([enhDist]))
                (E0P2, _), __ = self.tmmP2.layers[0].GetFields(np.array([0.0]), outBackAndForward=True)
                enhP2 = np.linalg.norm(EP2[0, :]) / (
                    np.linalg.norm(E0P2[0, :]) * np.sqrt(self.tmmP2.layers[0].n(self.wlP2).real)
                )

                res["enhP1"][i] = enhP1
                res["enhP2"][i] = enhP2

        return res

    def _SolveFundamentalFields(self) -> None:
        # Solve fundametal field 1
        self.tmmP1 = _NonlinearTmm(
            I0=self.I0P1, overrideE0=self.overrideE0P1, wl=self.wlP1, pol=self.polP1, mode=_NonlinearTmm.MODE_INCIDENT
        )

        for d, n, _ in self.layers:
            self.tmmP1.AddLayer(d, n)
        self.resP1 = self.tmmP1.Solve(beta=self.betaP1)

        # Solve fundametal field 2
        self.tmmP2 = _NonlinearTmm(
            I0=self.I0P2, overrideE0=self.overrideE0P2, wl=self.wlP2, pol=self.polP2, mode=_NonlinearTmm.MODE_INCIDENT
        )

        for d, n, _ in self.layers:
            self.tmmP2.AddLayer(d, n)
        self.resP2 = self.tmmP2.Solve(beta=self.betaP2)



from __future__ import annotations

import math

import matplotlib.pyplot as plt
import numpy as np

from NonlinearTMM import TMM, Material


def CalcSpp() -> None:
    # Parameters
    # ---------------------------------------------------------------------------
    wl = 532e-9  # Wavelength
    pol = "p"  # Polarization
    I0 = 1.0  # Intensity of incident wave
    metalD = 50e-9  # Metal film thickness
    enhLayer = 2  # Measure enhancment in the last layer
    ths = np.radians(np.linspace(0.0, 80.0, 500))  # Angle of incidences
    xs = np.linspace(-2e-6, 2e-6, 200)  # Field calculation coordinates
    zs = np.linspace(-2e-6, 2e-6, 201)  # Field calculation coordinates

    # Specify materials
    # ---------------------------------------------------------------------------
    prism = Material.Static(1.5)
    ag = Material.Static(0.054007 + 3.4290j)  # Johnson & Christie @ 532nm
    dielectric = Material.Static(1.0)

    # Init TMM
    # ---------------------------------------------------------------------------
    tmm = TMM(wl=wl, pol=pol, I0=I0)
    tmm.AddLayer(math.inf, prism)
    tmm.AddLayer(metalD, ag)
    tmm.AddLayer(math.inf, dielectric)

    # Solve
    # ---------------------------------------------------------------------------

    # Calculate reflection, transmission and field enhancement
    betas = np.sin(ths) * prism.GetN(wl).real
    sweepRes = tmm.Sweep("beta", betas, outEnh=True, layerNr=enhLayer)

    # Calculate fields at the reflection dip (excitation of SPPs)
    betaMaxEnh = betas[np.argmax(sweepRes.enh)]
    tmm.Solve(beta=betaMaxEnh)

    # Calculate 1D fields
    fields1D = tmm.GetFields(zs)

    # Calculate 2D fields
    fields2D = tmm.GetFields2D(zs, xs)

    # Ploting
    # ---------------------------------------------------------------------------
    plt.figure()
    thMaxEnh = np.arcsin(betaMaxEnh / prism.GetN(wl).real)

    # Reflection / transmission
    plt.subplot(221)
    plt.plot(np.degrees(ths), sweepRes.Ir, label="R")
    plt.plot(np.degrees(ths), sweepRes.It, label="T")
    plt.axvline(np.degrees(thMaxEnh), ls="--", color="red", lw=1.0)
    plt.xlabel(r"$\theta$ ($\degree$)")
    plt.ylabel(r"Intensity (a.u)")
    plt.legend()

    # Field enhancement
    plt.subplot(222)
    plt.plot(np.degrees(ths), sweepRes.enh)
    plt.axvline(np.degrees(thMaxEnh), ls="--", color="red", lw=1.0)
    plt.xlabel(r"$\theta$ ($\degree$)")
    plt.ylabel(r"Field enhancement")

    # Fields 1D
    plt.subplot(223)
    plt.plot(1e6 * zs, fields1D.E[:, 0].real, label=r"$E_x$")
    plt.plot(1e6 * zs, fields1D.E[:, 2].real, label=r"$E_z$")
    plt.plot(1e6 * zs, np.linalg.norm(fields1D.E, axis=1), label=r"‖E‖")
    plt.xlabel(r"z (μm)")
    plt.ylabel(r"(V/m)")
    plt.legend()

    # Fields 2D
    plt.subplot(224)
    assert fields2D.Ez is not None
    plt.pcolormesh(1e6 * zs, 1e6 * xs, fields2D.Ez.real.T, rasterized=True)
    plt.xlabel(r"z (μm)")
    plt.ylabel(r"x (μm)")
    plt.colorbar(label=r"$E_z$ (V/m)")

    plt.tight_layout()
    plt.savefig("docs/images/TMM-example.png", dpi=100)
    plt.show()


if __name__ == "__main__":
    CalcSpp()



"""Tests for Material constructors and basic error paths."""

from __future__ import annotations

import numpy as np
import pytest

from NonlinearTMM import TMM, Material

# ---------------------------------------------------------------------------
# Material.Static
# ---------------------------------------------------------------------------


class TestMaterialStatic:
    def test_real_refractive_index(self):
        mat = Material.Static(1.5)
        assert mat.GetN(532e-9) == pytest.approx(1.5)

    def test_complex_refractive_index(self):
        n = 0.054007 + 3.4290j
        mat = Material.Static(n)
        assert mat.GetN(532e-9) == pytest.approx(n)

    def test_wavelength_independent(self):
        mat = Material.Static(2.0)
        n400 = mat.GetN(400e-9)
        n800 = mat.GetN(800e-9)
        assert n400 == pytest.approx(n800)


# ---------------------------------------------------------------------------
# Material(wls, ns) constructor
# ---------------------------------------------------------------------------


class TestMaterialConstructor:
    def test_interpolation(self):
        wls = np.array([400e-9, 600e-9, 800e-9])
        ns = np.array([1.5, 1.4, 1.3], dtype=complex)
        mat = Material(wls, ns)
        # Interior point should be interpolated
        n_mid = mat.GetN(600e-9)
        assert n_mid == pytest.approx(1.4)
        # Value between endpoints
        n_500 = mat.GetN(500e-9)
        assert 1.3 < np.real(n_500) < 1.5

    def test_complex_ns(self):
        wls = np.array([400e-9, 800e-9])
        ns = np.array([1.5 + 0.1j, 1.3 + 0.2j], dtype=complex)
        mat = Material(wls, ns)
        n = mat.GetN(400e-9)
        assert np.real(n) == pytest.approx(1.5)
        assert np.imag(n) == pytest.approx(0.1)


# ---------------------------------------------------------------------------
# Error path tests
# ---------------------------------------------------------------------------


class TestTMMErrors:
    def test_invalid_polarization(self):
        with pytest.raises((ValueError, RuntimeError)):
            TMM(wl=532e-9, pol="x", I0=1.0)

    def test_sweep_no_layers(self):
        tmm = TMM(wl=532e-9, pol="p", I0=1.0)
        with pytest.raises((ValueError, RuntimeError)):
            betas = np.linspace(0, 1, 10)
            tmm.Sweep("beta", betas)


# ---------------------------------------------------------------------------
# Public API surface test
# ---------------------------------------------------------------------------


class TestPublicAPI:
    def test_all_exports_importable(self):
        import NonlinearTMM

        for name in NonlinearTMM.__all__:
            assert hasattr(NonlinearTMM, name), f"{name} listed in __all__ but not importable"



from __future__ import annotations

import math
from typing import Any

import LabPy
import numpy as np
import pytest

from NonlinearTMM import TMM, Material


@pytest.fixture(params=[400e-9, 800e-9])
def wl(request):
    return request.param


@pytest.fixture(params=[1.5, 2.0])
def nPrism(request):
    return request.param


@pytest.fixture(params=[1.0, 1.3])
def nDielectric(request):
    return request.param


@pytest.fixture(params=["p", "s"])
def pol(request):
    return request.param


@pytest.fixture(params=[50e-9, 100e-9])
def metalD(request):
    return request.param


@pytest.fixture(params=[0.1, 1.0])
def I0(request):
    return request.param


@pytest.fixture(params=[1.0, 1.0 + 3j, 1j])
def overrideE0(request):
    return request.param


@pytest.fixture(params=[0.0, 0.9, 1.4])
def betaTest(request):
    return request.param


@pytest.fixture()
def tmms1(request, wl, nPrism, nDielectric, pol, metalD, I0):
    prism = LabPy.Material("Static", n=nPrism)
    metal = LabPy.Material("main/Ag/Johnson")
    dielectric = LabPy.Material("Static", n=nDielectric)
    tmmCpp, tmmPy = GetTMMs(wl, pol, I0, prism, metal, dielectric, metalD, None)

    return tmmCpp, tmmPy


@pytest.fixture()
def tmms2(request, nPrism, nDielectric, pol, overrideE0):
    wl = 532e-9
    metalD = 50e-9
    I0 = 1.0

    prism = LabPy.Material("Static", n=nPrism)
    metal = LabPy.Material("main/Ag/Johnson")
    dielectric = LabPy.Material("Static", n=nDielectric)
    tmmCpp, tmmPy = GetTMMs(wl, pol, I0, prism, metal, dielectric, metalD, overrideE0)

    return tmmCpp, tmmPy


def test_Powerflows1(tmms1):
    TestPowerFlows(tmms1)


def test_Powerflows2(tmms2):
    TestPowerFlows(tmms2)


def test_Fields1D(tmms2, betaTest):
    tmmCpp, tmmPy = tmms2
    zs = np.linspace(-200e-9, 200e-9, 20)

    tmmCpp.Solve(beta=betaTest)
    tmmPy.Solve(beta=betaTest)

    resEPy, resHPy = tmmPy.GetFields(zs)
    resCpp = tmmCpp.GetFields(zs)

    np.testing.assert_allclose(resEPy, resCpp.E)
    np.testing.assert_allclose(resHPy, resCpp.H)


def test_Fields2D(tmms2, betaTest):
    tmmCpp, tmmPy = tmms2
    xs = np.linspace(-300e-9, 250e-9, 200)
    zs = np.linspace(-200e-9, 200e-9, 200)

    tmmCpp.Solve(beta=betaTest)
    tmmPy.Solve(beta=betaTest)

    resEPy, resHPy = tmmPy.GetFields2D(zs, xs)
    resCpp = tmmCpp.GetFields2D(zs, xs)

    if tmmPy.pol == "s":
        np.testing.assert_allclose(resEPy[:, :, 1], resCpp.Ey)
        np.testing.assert_allclose(resHPy[:, :, 0], resCpp.Hx)
        np.testing.assert_allclose(resHPy[:, :, 2], resCpp.Hz)
    else:
        np.testing.assert_allclose(resHPy[:, :, 1], resCpp.Hy)
        np.testing.assert_allclose(resEPy[:, :, 0], resCpp.Ex)
        np.testing.assert_allclose(resEPy[:, :, 2], resCpp.Ez)


# ------------------------------------------------------------------------------


def GetTMMs(
    wl: float,
    pol: str,
    I0: float,
    prism: LabPy.Material,
    metal: LabPy.Material,
    dielectric: LabPy.Material,
    metalD: float,
    overrideE0: complex | None,
) -> tuple[TMM, Any]:
    # C++ TMM
    tmmCpp = TMM()
    tmmCpp.SetParams(wl=wl, pol=pol, I0=I0)

    if overrideE0 is not None:
        tmmCpp.SetParams(overrideE0=True, E0=overrideE0)

    tmmCpp.AddLayer(math.inf, Material.FromLabPy(prism))
    tmmCpp.AddLayer(metalD, Material.FromLabPy(metal))
    tmmCpp.AddLayer(math.inf, Material.FromLabPy(dielectric))

    # Python TMM
    tmmPy = LabPy._Tmm._NonlinearTmm._NonlinearTmm(wl=wl, pol=pol, I0=I0, mode="incident", overrideE0=overrideE0)
    tmmPy.AddLayer(math.inf, prism)
    tmmPy.AddLayer(metalD, metal)
    tmmPy.AddLayer(math.inf, dielectric)
    tmmPy.SetParams(beta=0.0)

    return tmmCpp, tmmPy


def TestPowerFlows(tmms: tuple[TMM, Any]) -> None:
    tmmCpp, tmmPy = tmms

    betas = np.linspace(0.0, 1.49, 20)
    sr = tmmCpp.Sweep("beta", betas)
    sweepFunc = np.vectorize(lambda beta: tmmPy.Solve(beta=beta))
    inc, r, t, Ii, R, T, _ = sweepFunc(betas)

    # Comparison
    np.testing.assert_allclose(sr.inc, inc)
    np.testing.assert_allclose(sr.r, r)
    np.testing.assert_allclose(sr.t, t)
    np.testing.assert_allclose(sr.Ii, Ii)
    np.testing.assert_allclose(sr.Ir, R)
    np.testing.assert_allclose(sr.It, T, atol=1e-9 * np.mean(Ii))
